# flake8: noqa
import pkg_resources
import sys
import warnings

from .config import config
from .type_sys import type_system
from .type_sys.utils import list_logical_types, list_semantic_tags
from .utils import concat_columns, read_file
from .version import __version__

import woodwork.column_accessor
import woodwork.demo
import woodwork.table_accessor
from woodwork.accessor_utils import (
    get_invalid_schema_message,
    init_series,
    is_schema_valid,
)

if sys.version_info.major == 3 and sys.version_info.minor == 7:  # pragma: no cover
    warnings.warn(
        "Woodwork may not support Python 3.7 in next non-bugfix release.", FutureWarning
    )

# Call functions registered by other libraries when woodwork is imported
for entry_point in pkg_resources.iter_entry_points(
    "alteryx_open_src_initialize"
):  # pragma: no cover
    try:
        method = entry_point.load()
        if callable(method):
            method("woodwork")
    except Exception:
        pass



class DuplicateTagsWarning(UserWarning):
    def get_warning_message(self, duplicate_tags, name):
        return f"Semantic tag(s) '{', '.join(duplicate_tags)}' already present on column '{name}'"


class StandardTagsChangedWarning(UserWarning):
    def get_warning_message(self, use_standard_tags, col_name=None):
        changed = "added to" if use_standard_tags else "removed from"
        name = ('"' + col_name + '"') if col_name is not None else "your column"
        return f"Standard tags have been {changed} {name}"


class UpgradeSchemaWarning(UserWarning):
    def get_warning_message(self, saved_version_str, current_schema_version):
        return (
            "The schema version of the saved Woodwork table "
            "%s is greater than the latest supported %s. "
            "You may need to upgrade woodwork. Attempting to load Woodwork table ..."
            % (saved_version_str, current_schema_version)
        )


class OutdatedSchemaWarning(UserWarning):
    def get_warning_message(self, saved_version_str):
        return (
            "The schema version of the saved Woodwork table "
            "%s is no longer supported by this version "
            "of woodwork. Attempting to load Woodwork table ..." % (saved_version_str)
        )


class IndexTagRemovedWarning(UserWarning):
    pass


class TypingInfoMismatchWarning(UserWarning):
    def get_warning_message(self, attr, invalid_reason, object_type):
        return (
            f"Operation performed by {attr} has invalidated the Woodwork typing information:\n "
            f"{invalid_reason}.\n "
            f"Please initialize Woodwork with {object_type}.ww.init"
        )


class TypeConversionError(Exception):
    def __init__(self, series, new_dtype, logical_type):
        message = f"Error converting datatype for {series.name} from type {str(series.dtype)} "
        message += f"to type {new_dtype}. Please confirm the underlying data is consistent with "
        message += f"logical type {logical_type}."
        super().__init__(message)


class TypeConversionWarning(UserWarning):
    pass


class ParametersIgnoredWarning(UserWarning):
    pass


class ColumnNotPresentError(KeyError):
    def __init__(self, column):
        if isinstance(column, str):
            return super().__init__(
                f"Column with name '{column}' not found in DataFrame"
            )
        elif isinstance(column, list):
            return super().__init__(f"Column(s) '{column}' not found in DataFrame")


class ColumnNotPresentInSchemaError(KeyError):
    def __init__(self, column):
        return super().__init__(
            f"Column(s) {column} not found in Woodwork schema. Please initialize with DataFrame.ww.init"
        )


class WoodworkNotInitError(AttributeError):
    pass


class WoodworkNotInitWarning(UserWarning):
    pass


class TypeValidationError(Exception):
    pass


class SparseDataWarning(UserWarning):
    pass


class WoodworkFileExistsError(FileExistsError):
    pass



"""
All code in this module was copied from pandas and is subject to the following
license:

BSD 3-Clause License

Copyright (c) 2008-2011, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team
All rights reserved.

Copyright (c) 2011-2021, Open source contributors.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from dateutil.parser import parse as du_parse
from pandas._libs.tslibs.parsing import _timelex


def guess_datetime_format(
    dt_str,
    dayfirst=False,
    dt_str_parse=du_parse,
    dt_str_split=_timelex.split,
):  # pragma: no cover
    """
    TODO:

    This is a backport of the fix submitted to pandas in this PR:
    https://github.com/pandas-dev/pandas/pull/43900

    The usual pandas function is
    `pandas.core.tools.datetimes.guess_datetime_format`.  Once we upgrade to a
    version of pandas with the fix (probably v1.4), we should delete this
    backport and just use the pandas function.  We'll also need to delete the
    test for this backport in the test suite.
    """
    if dt_str_parse is None or dt_str_split is None:
        return None

    if not isinstance(dt_str, str):
        return None

    day_attribute_and_format = (("day",), "%d", 2)

    # attr name, format, padding (if any)
    datetime_attrs_to_format = [
        (("year", "month", "day"), "%Y%m%d", 0),
        (("year",), "%Y", 0),
        (("month",), "%B", 0),
        (("month",), "%b", 0),
        (("month",), "%m", 2),
        day_attribute_and_format,
        (("hour",), "%H", 2),
        (("minute",), "%M", 2),
        (("second",), "%S", 2),
        (("microsecond",), "%f", 6),
        (("second", "microsecond"), "%S.%f", 0),
        (("tzinfo",), "%z", 0),
        (("tzinfo",), "%Z", 0),
        (("day_of_week",), "%a", 0),
        (("day_of_week",), "%A", 0),
        (("meridiem",), "%p", 0),
    ]

    if dayfirst:
        datetime_attrs_to_format.remove(day_attribute_and_format)
        datetime_attrs_to_format.insert(0, day_attribute_and_format)

    try:
        parsed_datetime = dt_str_parse(dt_str, dayfirst=dayfirst)
    except (ValueError, OverflowError):
        # In case the datetime can't be parsed, its format cannot be guessed
        return None

    if parsed_datetime is None:
        return None

    # the default dt_str_split from dateutil will never raise here; we assume
    #  that any user-provided function will not either.
    tokens = dt_str_split(dt_str)

    # Normalize offset part of tokens.
    # There are multiple formats for the timezone offset.
    # To pass the comparison condition between the output of `strftime` and
    # joined tokens, which is carried out at the final step of the function,
    # the offset part of the tokens must match the '%z' format like '+0900'
    # instead of ‘+09:00’.
    if parsed_datetime.tzinfo is not None:
        offset_index = None
        if len(tokens) > 0 and tokens[-1] == "Z":
            # the last 'Z' means zero offset
            offset_index = -1
        elif len(tokens) > 1 and tokens[-2] in ("+", "-"):
            # ex. [..., '+', '0900']
            offset_index = -2
        elif len(tokens) > 3 and tokens[-4] in ("+", "-"):
            # ex. [..., '+', '09', ':', '00']
            offset_index = -4

        if offset_index is not None:
            # If the input string has a timezone offset like '+0900',
            # the offset is separated into two tokens, ex. ['+', '0900’].
            # This separation will prevent subsequent processing
            # from correctly parsing the time zone format.
            # So in addition to the format nomalization, we rejoin them here.
            tokens[offset_index] = parsed_datetime.strftime("%z")
            tokens = tokens[: offset_index + 1 or None]

    format_guess = [None] * len(tokens)
    found_attrs = set()

    for attrs, attr_format, padding in datetime_attrs_to_format:
        # If a given attribute has been placed in the format string, skip
        # over other formats for that same underlying attribute (IE, month
        # can be represented in multiple different ways)
        if set(attrs) & found_attrs:
            continue

        if parsed_datetime.tzinfo is None and attr_format in ("%Z", "%z"):
            continue

        parsed_formatted = parsed_datetime.strftime(attr_format)
        for i, token_format in enumerate(format_guess):
            token_filled = tokens[i].zfill(padding)
            if token_format is None and token_filled == parsed_formatted:
                format_guess[i] = attr_format
                tokens[i] = token_filled
                found_attrs.update(attrs)
                break

    # Only consider it a valid guess if we have a year, month and day
    if len({"year", "month", "day"} & found_attrs) != 3:
        return None

    output_format = []
    for i, guess in enumerate(format_guess):
        if guess is not None:
            # Either fill in the format placeholder (like %Y)
            output_format.append(guess)
        else:
            # Or just the token separate (IE, the dashes in "01-01-2013")
            try:
                # If the token is numeric, then we likely didn't parse it
                # properly, so our guess is wrong
                float(tokens[i])
                return None
            except ValueError:
                pass

            output_format.append(tokens[i])

    guessed_format = "".join(output_format)

    # rebuild string, capturing any inferred padding
    dt_str = "".join(tokens)
    if parsed_datetime.strftime(guessed_format) == dt_str:
        return guessed_format
    else:
        return None



import copy
import warnings
import weakref
from typing import Any, Callable, Dict, Iterable, List, Sequence, Set, Union

import pandas as pd

from woodwork.accessor_utils import (
    _check_table_schema,
    _is_dask_dataframe,
    _is_dataframe,
    _is_spark_dataframe,
    get_invalid_schema_message,
    init_series,
)
from woodwork.exceptions import (
    ColumnNotPresentError,
    IndexTagRemovedWarning,
    ParametersIgnoredWarning,
    TypingInfoMismatchWarning,
)
from woodwork.indexers import _iLocIndexer, _locIndexer
from woodwork.logical_types import Datetime, LogicalType
from woodwork.serializers.serializer_base import typing_info_to_dict
from woodwork.serializers.utils import get_serializer
from woodwork.statistics_utils import (
    _get_dependence_dict,
    _get_describe_dict,
    _get_valid_mi_columns,
    _get_value_counts,
    _infer_temporal_frequencies,
)
from woodwork.table_schema import TableSchema
from woodwork.type_sys.utils import _is_numeric_series, col_is_datetime
from woodwork.typing import AnyDataFrame, ColumnName, UseStandardTagsDict
from woodwork.utils import _get_column_logical_type, _parse_logical_type, import_or_none

dd = import_or_none("dask.dataframe")
ps = import_or_none("pyspark.pandas")


class WoodworkTableAccessor:
    def __init__(self, dataframe):
        self._dataframe_weakref = weakref.ref(dataframe)
        self._schema = None

    def init(self, **kwargs):
        """Initializes Woodwork typing information for a DataFrame with a partial schema.

        Logical type priority:
            1. Types specified in ``logical_types``
            2. Types specified in ``partial_schema``
            3. Types inferred by ``ww.type_system.infer_logical_type``

        Other Info priority:
            1. Parameter passed in
            2. Value specified in ``partial_schema``

        Args:
            schema (Woodwork.TableSchema, optional): Typing information to use for the DataFrame instead of performing inference.
                 Specified arguments will override the schema's typing information.
            index (str, optional): Name of the index column.
            time_index (str, optional): Name of the time index column.
            logical_types (Dict[str -> LogicalType], optional): Dictionary mapping column names in
                the DataFrame to the LogicalType for the column. Setting a column's logical type to None in this dict will
                force a logical to be inferred.
            already_sorted (bool, optional): Indicates whether the input DataFrame is already sorted on the time
                index. If False, will sort the dataframe first on the time_index and then on the index (pandas DataFrame
                only). Defaults to False.
            name (str, optional): Name used to identify the DataFrame.
            semantic_tags (dict, optional): Dictionary mapping column names in Woodwork to the
                semantic tags for the column. The keys in the dictionary should be strings
                that correspond to column names. There are two options for
                specifying the dictionary values:
                (str): If only one semantic tag is being set, a single string can be used as a value.
                (list[str] or set[str]): If multiple tags are being set, a list or set of strings can be
                used as the value.
                Semantic tags will be set to an empty set for any column not included in the
                dictionary.
            table_metadata (Dict[str -> json serializable], optional): Dictionary containing extra metadata for Woodwork.
            column_metadata (Dict[str -> Dict[str -> json serializable]], optional): Dictionary mapping column names
                to that column's metadata dictionary.
            use_standard_tags (bool, Dict[str -> bool], optional): Determines whether standard semantic tags will be
                added to columns based on the specified logical type for the column.
                If a single boolean is supplied, will apply the same use_standard_tags value to all columns.
                A dictionary can be used to specify ``use_standard_tags`` values for individual columns.
                Unspecified columns will use the default value of True.
            column_descriptions (Dict[str -> str], optional): Dictionary mapping column names to column descriptions.
            column_origins (str, Dict[str -> str], optional): Origin of each column. If a string is supplied, it is
                used as the origin for all columns. A dictionary can be used to set origins for individual columns.
            validate (bool, optional): Whether parameter and data validation should occur. Defaults to True. Warning:
                Should be set to False only when parameters and data are known to be valid.
                Any errors resulting from skipping validation with invalid inputs may not be easily understood.
        """
        self.init_with_partial_schema(**kwargs)

    def init_with_full_schema(
        self, schema: TableSchema, validate: bool = True, **kwargs
    ) -> None:
        """Initializes Woodwork typing information for a DataFrame with a complete schema.

        Args:
            schema (Woodwork.TableSchema): Typing information to use for the DataFrame instead of performing inference. Note that any changes made to the
                schema object after initialization will propagate to the DataFrame. Similarly, to avoid unintended typing information changes, the same schema
                object should not be shared between DataFrames.
            validate (bool, optional): Whether parameter and data validation should occur. Defaults to True. Warning: Should be set to False only when
                parameters and data are known to be valid. Any errors resulting from skipping validation with invalid inputs may not be easily understood.
        """
        if validate:
            _check_schema(self._dataframe, schema)
            _check_unique_column_names(self._dataframe)
        self._schema = schema

        extra_params = [key for key in kwargs]
        if extra_params:
            warnings.warn(
                "A schema was provided and the following parameters were ignored: "
                + ", ".join(extra_params),
                ParametersIgnoredWarning,
            )

    def init_with_partial_schema(
        self,
        schema: TableSchema = None,
        index: str = None,
        time_index: str = None,
        logical_types: Dict[ColumnName, Union[str, LogicalType, None]] = None,
        already_sorted: bool = False,
        name: str = None,
        semantic_tags: Dict[ColumnName, Union[str, List[str], Set[str]]] = None,
        table_metadata: dict = None,
        column_metadata: Dict[ColumnName, dict] = None,
        use_standard_tags: Union[bool, UseStandardTagsDict] = None,
        column_descriptions: Dict[ColumnName, str] = None,
        column_origins: Union[str, Dict[ColumnName, str]] = None,
        validate: bool = True,
        **kwargs,
    ) -> None:
        """Initializes Woodwork typing information for a DataFrame with a partial schema.

        Logical type priority:
            1. Types specified in ``logical_types``
            2. Types specified in ``partial_schema``
            3. Types inferred by ``ww.type_system.infer_logical_type``

        Other Info priority:
            1. Parameter passed in
            2. Value specified in ``partial_schema``

        Args:
            schema (Woodwork.TableSchema, optional): Typing information to use for the DataFrame instead of performing inference.
                 Specified arguments will override the schema's typing information.
            index (str, optional): Name of the index column.
            time_index (str, optional): Name of the time index column.
            logical_types (Dict[str -> LogicalType], optional): Dictionary mapping column names in
                the DataFrame to the LogicalType for the column. Setting a column's logical type to None in this dict will
                force a logical to be inferred.
            already_sorted (bool, optional): Indicates whether the input DataFrame is already sorted on the time
                index. If False, will sort the dataframe first on the time_index and then on the index (pandas DataFrame
                only). Defaults to False.
            name (str, optional): Name used to identify the DataFrame.
            semantic_tags (dict, optional): Dictionary mapping column names in Woodwork to the
                semantic tags for the column. The keys in the dictionary should be strings
                that correspond to column names. There are two options for
                specifying the dictionary values:
                (str): If only one semantic tag is being set, a single string can be used as a value.
                (list[str] or set[str]): If multiple tags are being set, a list or set of strings can be
                used as the value.
                Semantic tags will be set to an empty set for any column not included in the
                dictionary.
            table_metadata (Dict[str -> json serializable], optional): Dictionary containing extra metadata for Woodwork.
            column_metadata (Dict[str -> Dict[str -> json serializable]], optional): Dictionary mapping column names
                to that column's metadata dictionary.
            use_standard_tags (bool, Dict[str -> bool], optional): Determines whether standard semantic tags will be
                added to columns based on the specified logical type for the column.
                If a single boolean is supplied, will apply the same use_standard_tags value to all columns.
                A dictionary can be used to specify ``use_standard_tags`` values for individual columns.
                Unspecified columns will use the default value of True.
            column_descriptions (Dict[str -> str], optional): Dictionary mapping column names to column descriptions.
            column_origins (str, Dict[str -> str], optional): Origin of each column. If a string is supplied, it is
                used as the origin for all columns. A dictionary can be used to set origins for individual columns.
            validate (bool, optional): Whether parameter and data validation should occur. Defaults to True. Warning:
                Should be set to False only when parameters and data are known to be valid.
                Any errors resulting from skipping validation with invalid inputs may not be easily understood.
        """
        if validate:
            _validate_accessor_params(
                self._dataframe,
                index,
                time_index,
                logical_types,
                schema,
                use_standard_tags,
            )

        existing_logical_types = {}
        existing_col_descriptions = {}
        existing_col_metadata = {}
        existing_use_standard_tags = {}
        existing_semantic_tags = {}
        existing_col_origins = {}

        if schema:  # pull schema parameters
            name = name if name is not None else schema.name
            index = index if index is not None else schema.index
            time_index = time_index if time_index is not None else schema.time_index
            table_metadata = table_metadata or schema.metadata
            for col_name, col_schema in schema.columns.items():
                existing_logical_types[col_name] = col_schema.logical_type
                existing_semantic_tags[col_name] = (
                    col_schema.semantic_tags
                    - {"time_index"}
                    - {"index"}
                    - col_schema.logical_type.standard_tags
                )
                existing_col_descriptions[col_name] = col_schema.description
                existing_col_origins[col_name] = col_schema.origin
                existing_col_metadata[col_name] = col_schema.metadata
                existing_use_standard_tags[col_name] = col_schema.use_standard_tags

        # overwrite schema parameters with specified kwargs
        logical_types = _infer_missing_logical_types(
            self._dataframe, logical_types, existing_logical_types
        )
        column_descriptions = {
            **existing_col_descriptions,
            **(column_descriptions or {}),
        }
        column_metadata = {**existing_col_metadata, **(column_metadata or {})}
        column_names = list(self._dataframe.columns)
        use_standard_tags = _merge_use_standard_tags(
            existing_use_standard_tags, use_standard_tags, column_names
        )
        semantic_tags = {**existing_semantic_tags, **(semantic_tags or {})}
        column_origins = {**existing_col_origins, **(column_origins or {})}

        self._schema = TableSchema(
            column_names=column_names,
            logical_types=logical_types,
            name=name,
            index=index,
            time_index=time_index,
            semantic_tags=copy.deepcopy(semantic_tags),
            table_metadata=copy.deepcopy(table_metadata),
            column_metadata=copy.deepcopy(column_metadata),
            use_standard_tags=copy.deepcopy(use_standard_tags),
            column_descriptions=column_descriptions,
            column_origins=column_origins,
            validate=validate,
            **kwargs,
        )
        self._set_underlying_index()
        if self._schema.time_index is not None:
            self._sort_columns(already_sorted)

    def __eq__(self, other, deep=True):
        if not self._schema.__eq__(other.ww._schema, deep=deep):
            return False

        # Only check pandas DataFrames for equality
        if (
            deep
            and isinstance(self._dataframe, pd.DataFrame)
            and isinstance(other.ww._dataframe, pd.DataFrame)
        ):
            return self._dataframe.equals(other.ww._dataframe)
        return True

    @_check_table_schema
    def __getattr__(self, attr):
        # Called if method is not present on the Accessor
        # If the method is present on TableSchema, uses that method.
        # If the method is present on DataFrame, uses that method.
        if hasattr(self._schema, attr):
            return self._make_schema_call(attr)
        if hasattr(self._dataframe, attr):
            return self._make_dataframe_call(attr)
        else:
            raise AttributeError(f"Woodwork has no attribute '{attr}'")

    @_check_table_schema
    def __getitem__(self, key):
        if isinstance(key, list):
            columns = set(self._dataframe.columns)
            diff = list(set(key).difference(columns))

            if diff:
                raise ColumnNotPresentError(sorted(diff))

            return self._get_subset_df_with_schema(key)

        if key not in self._dataframe:
            raise ColumnNotPresentError(key)

        series = self._dataframe[key]
        column = copy.deepcopy(self._schema.columns[key])

        series.ww.init(schema=column, validate=False)

        return series

    def __setitem__(self, col_name, column):
        series = tuple(pkg.Series for pkg in (pd, dd, ps) if pkg)
        if not isinstance(column, series):
            raise ValueError("New column must be of Series type")

        if column.ww.schema is not None and "index" in column.ww.semantic_tags:
            warnings.warn(
                f'Cannot add "index" tag on {col_name} directly to the DataFrame. The "index" tag has been removed from {col_name}. To set this column as a Woodwork index, please use df.ww.set_index',
                IndexTagRemovedWarning,
            )
            column.ww.set_semantic_tags(column.ww.semantic_tags - {"index"})

        # Don't allow reassigning of index or time index with setitem
        if self.index == col_name:
            raise KeyError(
                "Cannot reassign index. Change column name and then use df.ww.set_index to reassign index."
            )
        if self.time_index == col_name:
            raise KeyError(
                "Cannot reassign time index. Change column name and then use df.ww.set_time_index to reassign time index."
            )

        if column.ww._schema is None:
            column = init_series(column, use_standard_tags=True)

        self._dataframe[col_name] = column
        self._schema.columns[col_name] = column.ww._schema

    def __repr__(self):
        """A string representation of a Woodwork table containing typing information"""
        return repr(self._get_typing_info())

    def _repr_html_(self):
        """An HTML representation of a Woodwork table for IPython.display in Jupyter Notebooks
        containing typing information and a preview of the data."""
        return self._get_typing_info().to_html()

    @_check_table_schema
    def _get_typing_info(self):
        """Creates a DataFrame that contains the typing information for a Woodwork table."""
        typing_info = self._schema._get_typing_info().copy()
        typing_info.insert(0, "Physical Type", pd.Series(self.physical_types))
        # Maintain the same column order used in the DataFrame
        typing_info = typing_info.loc[list(self._dataframe.columns), :]
        return typing_info

    @property
    @_check_table_schema
    def name(self):
        """Name of the DataFrame"""
        return self._schema.name

    @name.setter
    @_check_table_schema
    def name(self, name):
        """Set name of the DataFrame"""
        self._schema.name = name

    @property
    @_check_table_schema
    def metadata(self):
        """Metadata of the DataFrame"""
        return self._schema.metadata

    @metadata.setter
    @_check_table_schema
    def metadata(self, metadata):
        """Set metadata of the DataFrame"""
        self._schema.metadata = metadata

    @property
    def _dataframe(self):
        return self._dataframe_weakref()

    @property
    @_check_table_schema
    def iloc(self):
        """
        Integer-location based indexing for selection by position.
        ``.iloc[]`` is primarily integer position based (from ``0`` to
        ``length-1`` of the axis), but may also be used with a boolean array.

        If the selection result is a DataFrame or Series, Woodwork typing
        information will be initialized for the returned object when possible.

        Allowed inputs are:
            An integer, e.g. ``5``.
            A list or array of integers, e.g. ``[4, 3, 0]``.
            A slice object with ints, e.g. ``1:7``.
            A boolean array.
            A ``callable`` function with one argument (the calling Series, DataFrame
            or Panel) and that returns valid output for indexing (one of the above).
            This is useful in method chains, when you don't have a reference to the
            calling object, but would like to base your selection on some value.
        """
        return _iLocIndexer(self._dataframe)

    @property
    @_check_table_schema
    def loc(self):
        """
        Access a group of rows by label(s) or a boolean array.

        ``.loc[]`` is primarily label based, but may also be used with a
        boolean array.

        If the selection result is a DataFrame or Series, Woodwork typing
        information will be initialized for the returned object when possible.

        Allowed inputs are:
            A single label, e.g. ``5`` or ``'a'``, (note that ``5`` is
            interpreted as a *label* of the index, and **never** as an
            integer position along the index).
            A list or array of labels, e.g. ``['a', 'b', 'c']``.
            A slice object with labels, e.g. ``'a':'f'``.
            A boolean array of the same length as the axis being sliced,
            e.g. ``[True, False, True]``.
            An alignable boolean Series. The index of the key will be aligned before
            masking.
            An alignable Index. The Index of the returned selection will be the input.
            A ``callable`` function with one argument (the calling Series or
            DataFrame) and that returns valid output for indexing (one of the above)
        """
        return _locIndexer(self._dataframe)

    @property
    def schema(self):
        """A copy of the Woodwork typing information for the DataFrame."""
        if self._schema:
            return copy.deepcopy(self._schema)

    @property
    @_check_table_schema
    def physical_types(self):
        """A dictionary containing physical types for each column"""
        return {
            col_name: self._schema.logical_types[col_name]._get_valid_dtype(
                type(self._dataframe[col_name])
            )
            for col_name in self._dataframe.columns
        }

    @property
    @_check_table_schema
    def types(self):
        """DataFrame containing the physical dtypes, logical types and semantic
        tags for the schema."""
        return self._get_typing_info()

    @property
    @_check_table_schema
    def logical_types(self):
        """A dictionary containing logical types for each column"""
        return self._schema.logical_types

    @property
    @_check_table_schema
    def semantic_tags(self):
        """A dictionary containing semantic tags for each column"""
        return self._schema.semantic_tags

    @property
    @_check_table_schema
    def index(self):
        """The index column for the table"""
        return self._schema.index

    @property
    @_check_table_schema
    def time_index(self):
        """The time index column for the table"""
        return self._schema.time_index

    @property
    @_check_table_schema
    def use_standard_tags(self):
        """A dictionary containing the use_standard_tags setting for each column in the table"""
        return self._schema.use_standard_tags

    @_check_table_schema
    def set_index(self, new_index):
        """Sets the index column of the DataFrame. Adds the 'index' semantic tag to the column
        and clears the tag from any previously set index column.

        Setting a column as the index column will also cause any previously set standard
        tags for the column to be removed.

        Clears the DataFrame's index by passing in None.

        Args:
            new_index (str): The name of the column to set as the index
        """
        self._schema.set_index(new_index)

        if self._schema.index is not None:
            _check_index(self._dataframe, self._schema.index)
        self._set_underlying_index()

    @_check_table_schema
    def set_time_index(self, new_time_index):
        """Set the time index. Adds the 'time_index' semantic tag to the column and
        clears the tag from any previously set index column

        Args:
            new_time_index (str): The name of the column to set as the time index.
                If None, will remove the time_index.
        """
        self._schema.set_time_index(new_time_index)

    @_check_table_schema
    def set_types(self, logical_types=None, semantic_tags=None, retain_index_tags=True):
        """Update the logical type and semantic tags for any columns names in the provided types dictionaries,
        updating the Woodwork typing information for the DataFrame.

        Args:
            logical_types (Dict[str -> str], optional): A dictionary defining the new logical types for the
                specified columns.
            semantic_tags (Dict[str -> str/list/set], optional): A dictionary defining the new semantic_tags for the
                specified columns.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will replace all semantic tags any time a column's
                semantic tags or logical type changes. Defaults to True.
        """
        logical_types = logical_types or {}
        logical_types = {
            col_name: _parse_logical_type(ltype, col_name)
            for col_name, ltype in logical_types.items()
        }

        self._schema.set_types(
            logical_types=logical_types,
            semantic_tags=semantic_tags,
            retain_index_tags=retain_index_tags,
        )
        # go through changed ltypes and update dtype if necessary
        for col_name, logical_type in logical_types.items():
            series = self._dataframe[col_name]
            updated_series = logical_type.transform(series)
            if updated_series is not series:
                self._dataframe[col_name] = updated_series

    @_check_table_schema
    def select(self, include=None, exclude=None, return_schema=False):
        """Create a DataFrame with Woodwork typing information initialized
        that includes only columns whose Logical Type and semantic tags match
        conditions specified in the list of types and tags to include or exclude.
        Values for both ``include`` and ``exclude`` cannot be provided in a
        single call.

        If no matching columns are found, an empty DataFrame will be returned.

        Args:
            include (str or LogicalType or list[str or LogicalType]): Logical
                types, semantic tags to include in the DataFrame.
            exclude (str or LogicalType or list[str or LogicalType]): Logical
                types, semantic tags to exclude from the DataFrame.
            return_schema (boolen): If True, return only the schema for the
                matching columns. Defaults to False

        Returns:
            DataFrame: The subset of the original DataFrame that matches the
            conditions specified by ``include`` or ``exclude``. Has Woodwork
            typing information initialized.
        """
        if include is not None and exclude is not None:
            raise ValueError(
                "Cannot specify values for both 'include' and 'exclude' in a single call."
            )
        if include is None and exclude is None:
            raise ValueError("Must specify values for either 'include' or 'exclude'.")

        cols_to_include = self._schema._filter_cols(include, exclude)

        if return_schema:
            return self._schema.get_subset_schema(cols_to_include)
        return self._get_subset_df_with_schema(cols_to_include)

    @_check_table_schema
    def add_semantic_tags(self, semantic_tags):
        """Adds specified semantic tags to columns, updating the Woodwork typing information.
        Will retain any previously set values.

        Args:
            semantic_tags (Dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataFrame to the tags that should be added to the column's semantic tags
        """
        self._schema.add_semantic_tags(semantic_tags)

    @_check_table_schema
    def remove_semantic_tags(self, semantic_tags):
        """Remove the semantic tags for any column names in the provided semantic_tags
        dictionary, updating the Woodwork typing information. Including `index` or `time_index`
        tags will set the Woodwork index or time index to None for the DataFrame.

        Args:
            semantic_tags (Dict[str -> str/list/set]): A dictionary mapping the columns
                in the DataFrame to the tags that should be removed from the column's semantic tags
        """
        self._schema.remove_semantic_tags(semantic_tags)

    @_check_table_schema
    def reset_semantic_tags(self, columns=None, retain_index_tags=False):
        """Reset the semantic tags for the specified columns to the default values.
        The default values will be either an empty set or a set of the standard tags
        based on the column logical type, controlled by the use_standard_tags property on each column.
        Column names can be provided as a single string, a list of strings or a set of strings.
        If columns is not specified, tags will be reset for all columns.

        Args:
            columns (str/list/set, optional): The columns for which the semantic tags should be reset.
            retain_index_tags (bool, optional): If True, will retain any index or time_index
                semantic tags set on the column. If False, will clear all semantic tags. Defaults to
                False.
        """
        self._schema.reset_semantic_tags(
            columns=columns, retain_index_tags=retain_index_tags
        )

    @_check_table_schema
    def to_dictionary(self):
        """Get a dictionary representation of the Woodwork typing information.

        Returns:
            dict: Description of the typing information.
        """
        return typing_info_to_dict(self._dataframe)

    @_check_table_schema
    def to_disk(
        self,
        path,
        format="csv",
        filename=None,
        data_subdirectory="data",
        typing_info_filename="woodwork_typing_info.json",
        compression=None,
        profile_name=None,
        **kwargs,
    ):
        """Write Woodwork table to disk in the format specified by `format`, location specified by `path`.
        Path could be a local path or an S3 path.
        If writing to S3 a tar archive of files will be written.

        Note:
            When serializing to parquet, Woodwork typing information will be stored in the parquet file metadata,
            and not in a separate typing info JSON file. Any value provided for `typing_info_filename` will be ignored.

        Args:
            path (str): Location on disk to write to (will be created as a directory if it does not exist)
            format (str, optional): Format to use for writing Woodwork data. Defaults to csv. Possible values are: {'csv', 'pickle', 'parquet'}.
            filename (str, optional): Name to use for the saved data file. Will default to the name of the dataframe or "data" if not specified.
            data_subdirectory (str, optional): Optional subdirectory to append to `path`. Will default to "data" if not specified.
            typing_info_filename (str, optional): Optional filename to use for storing Woodwork typing information JSON data.
                Will default to "woodwork_typing_info.json" if not specified. Will be ignored if serializing to parquet.
            compression (str, optional): Name of the compression to use. Possible values are: {'gzip', 'bz2', 'zip', 'xz', None}. Defaults to None.
            profile_name (str, optional): Name of AWS profile to use, False to use an anonymous profile, or None. Defaults to None.
            kwargs (keywords, optional): Additional keyword arguments to pass as keywords arguments to the underlying serialization method or to specify AWS profile.
        """
        serializer_cls = get_serializer(format)
        serializer = serializer_cls(
            path=path,
            filename=filename,
            data_subdirectory=data_subdirectory,
            typing_info_filename=typing_info_filename,
        )
        serializer.serialize(
            self._dataframe,
            compression=compression,
            profile_name=profile_name,
            **kwargs,
        )

    def _sort_columns(self, already_sorted):
        if _is_dask_dataframe(self._dataframe) or _is_spark_dataframe(self._dataframe):
            already_sorted = True  # Skip sorting for Dask and Spark input
        if not already_sorted:
            sort_cols = [self._schema.time_index, self._schema.index]
            if self._schema.index is None:
                sort_cols = [self._schema.time_index]
            self._dataframe.sort_values(sort_cols, inplace=True)

    def _set_underlying_index(self):
        """Sets the index of the underlying DataFrame to match the index column as
        specified by the TableSchema. Does not change the underlying index if no Woodwork index is
        specified. Only sets underlying index for pandas DataFrames.
        """
        if isinstance(self._dataframe, pd.DataFrame) and self._schema.index is not None:
            self._dataframe.set_index(self._schema.index, drop=False, inplace=True)
            # Drop index name to not overlap with the original column
            self._dataframe.index.name = None

    def _make_schema_call(self, attr):
        """Forwards the requested attribute onto the schema object.
        Results are that of the Woodwork.TableSchema class."""
        schema_attr = getattr(self._schema, attr)

        if callable(schema_attr):

            def wrapper(*args, **kwargs):
                return schema_attr(*args, **kwargs)

            return wrapper
        return schema_attr

    def _make_dataframe_call(self, attr):
        """Forwards the requested attribute onto the dataframe object. Intercepts return value,
        attempting to initialize Woodwork with the current schema when a new DataFrame is returned.
        Confirms schema is still valid for the original DataFrame."""
        dataframe_attr = getattr(self._dataframe, attr)

        if callable(dataframe_attr):

            def wrapper(*args, **kwargs):
                # Make DataFrame call and intercept the result
                result = dataframe_attr(*args, **kwargs)

                # Try to initialize Woodwork with the existing schema
                if _is_dataframe(result):
                    invalid_schema_message = get_invalid_schema_message(
                        result, self._schema
                    )
                    if invalid_schema_message:
                        warnings.warn(
                            TypingInfoMismatchWarning().get_warning_message(
                                attr, invalid_schema_message, "DataFrame"
                            ),
                            TypingInfoMismatchWarning,
                        )
                    else:
                        copied_schema = self.schema
                        result.ww.init_with_full_schema(
                            schema=copied_schema, validate=False
                        )
                else:
                    # Confirm that the schema is still valid on original DataFrame
                    # Important for inplace operations
                    invalid_schema_message = get_invalid_schema_message(
                        self._dataframe, self._schema
                    )

                    if invalid_schema_message:
                        warnings.warn(
                            TypingInfoMismatchWarning().get_warning_message(
                                attr, invalid_schema_message, "DataFrame"
                            ),
                            TypingInfoMismatchWarning,
                        )
                        self._schema = None

                # Always return the results of the DataFrame operation whether or not Woodwork is initialized
                return result

            return wrapper
        # Directly return non-callable DataFrame attributes
        return dataframe_attr

    def _get_subset_df_with_schema(self, cols_to_include, inplace=False):
        """Creates a new DataFrame from a list of column names with Woodwork initialized,
        retaining all typing information and maintaining the DataFrame's column order."""
        if inplace:
            if _is_dask_dataframe(self._dataframe):
                raise ValueError("Drop inplace not supported for Dask")
            if _is_spark_dataframe(self._dataframe):
                raise ValueError("Drop inplace not supported for Spark")

        assert all([col_name in self._schema.columns for col_name in cols_to_include])

        new_schema = self._schema.get_subset_schema(cols_to_include)
        if inplace:
            cols_to_drop = [
                col_name
                for col_name in self._dataframe.columns
                if col_name not in cols_to_include
            ]
            self._dataframe.drop(cols_to_drop, axis="columns", inplace=True)
            self.init_with_full_schema(schema=new_schema, validate=False)
            return
        new_df = self._dataframe[cols_to_include]
        new_df.ww.init_with_full_schema(schema=new_schema, validate=False)

        return new_df

    @_check_table_schema
    def pop(self, column_name):
        """Return a Series with Woodwork typing information and remove it from the DataFrame.

        Args:
            column (str): Name of the column to pop.

        Returns:
            Series: Popped series with Woodwork initialized
        """
        if column_name not in self._dataframe.columns:
            raise ColumnNotPresentError(column_name)

        series = self._dataframe.pop(column_name)

        # Initialize Woodwork typing info for series
        series.ww.init(schema=self.schema.columns[column_name], validate=False)

        # Update schema to not include popped column
        del self._schema.columns[column_name]

        return series

    @_check_table_schema
    def drop(self, columns, inplace=False):
        """Drop specified columns from a DataFrame.

        Args:
            columns (str or list[str]): Column name or names to drop. Must be present in the DataFrame.
            inplace (bool): If False, return a copy. Otherwise, do operation inplace and return None.

        Returns:
            DataFrame or None: DataFrame with the specified columns removed, maintaining Woodwork typing information or None if inplace=True. Only possible for pandas dataframes.

        Note:
            This method is used for removing columns only. To remove rows with ``drop``, go through the
            DataFrame directly and then reinitialize Woodwork with ``DataFrame.ww.init``
            instead of calling ``DataFrame.ww.drop``.
        """
        if not isinstance(columns, (list, set)):
            columns = [columns]

        not_present = [col for col in columns if col not in self._dataframe.columns]
        if not_present:
            raise ColumnNotPresentError(not_present)
        return self._get_subset_df_with_schema(
            [col for col in self._dataframe.columns if col not in columns],
            inplace=inplace,
        )

    @_check_table_schema
    def rename(self, columns, inplace=False):
        """Renames columns in a DataFrame, maintaining Woodwork typing information.

        Args:
            columns (Dict[str -> str]): A dictionary mapping current column names to new column names.
            inplace (bool): If False, return a copy. Otherwise, do operation inplace and return None.

        Returns:
            DataFrame or None: DataFrame with the specified columns renamed, maintaining Woodwork typing information or None if inplace=True. Only possible for pandas dataframes.
        """

        new_schema = self._schema.rename(columns)
        if inplace:
            if _is_dask_dataframe(self._dataframe):
                raise ValueError("Rename inplace not supported for Dask")
            if _is_spark_dataframe(self._dataframe):
                raise ValueError("Rename inplace not supported for Spark")
            self._dataframe.rename(columns=columns, inplace=True)
            self.init_with_full_schema(schema=new_schema)
            return
        new_df = self._dataframe.rename(columns=columns)

        new_df.ww.init_with_full_schema(schema=new_schema, validate=False)
        return new_df

    @_check_table_schema
    def mutual_information_dict(
        self,
        num_bins=10,
        nrows=None,
        include_index=False,
        callback=None,
        extra_stats=False,
        min_shared=25,
        random_seed=0,
    ):
        """
        Calculates mutual information between all pairs of columns in the DataFrame that
        support mutual information. Call woodwork.utils.get_valid_mi_types to see which Logical Types support
        mutual information.

        Args:
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical. Defaults to 10.
            nrows (int): The number of rows to sample for when determining mutual info.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for mutual information calculations.
                If False, the index column will not have mutual information calculated for it.
                Defaults to False.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:
                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            extra_stats (bool):  If True, additional column "shared_rows"
                recording the number of shared non-null rows for a column
                pair will be included with the dataframe. Defaults to False.
            min_shared (int): The number of shared non-null rows needed to
                calculate. Less rows than this will be considered too sparse
                to measure accurately and will return a NaN value. Must be
                non-negative. Defaults to 25.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        Returns:
            list(dict): A list containing dictionaries that have keys `column_1`,
            `column_2`, and `mutual_info` that is sorted in decending order by mutual info.
            Mutual information values are between 0 (no mutual information) and 1
            (perfect dependency).
        """
        return _get_dependence_dict(
            dataframe=self._dataframe,
            measures=["mutual_info"],
            num_bins=num_bins,
            nrows=nrows,
            include_index=include_index,
            callback=callback,
            extra_stats=extra_stats,
            min_shared=min_shared,
            random_seed=random_seed,
        )

    def mutual_information(
        self,
        num_bins=10,
        nrows=None,
        include_index=False,
        callback=None,
        extra_stats=False,
        min_shared=25,
        random_seed=0,
    ):
        """Calculates mutual information between all pairs of columns in the DataFrame that
        support mutual information. Call woodwork.utils.get_valid_mi_types to see which Logical Types support
        mutual information.

        Args:
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical. Defaults to 10.
            nrows (int): The number of rows to sample for when determining mutual info.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for mutual information calculations.
                If False, the index column will not have mutual information calculated for it.
                Defaults to False.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:
                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            extra_stats (bool):  If True, additional column "shared_rows"
                recording the number of shared non-null rows for a column
                pair will be included with the dataframe. Defaults to False.
            min_shared (int): The number of shared non-null rows needed to
                calculate. Less rows than this will be considered too sparse
                to measure accurately and will return a NaN value. Must be
                non-negative. Defaults to 25.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        Returns:
            pd.DataFrame: A DataFrame containing mutual information with columns `column_1`,
            `column_2`, and `mutual_info` that is sorted in decending order by mutual info.
            Mutual information values are between 0 (no mutual information) and 1
            (perfect dependency).
        """
        mutual_info = self.mutual_information_dict(
            num_bins=num_bins,
            nrows=nrows,
            include_index=include_index,
            callback=callback,
            extra_stats=extra_stats,
            min_shared=min_shared,
            random_seed=random_seed,
        )
        return pd.DataFrame(mutual_info)

    @_check_table_schema
    def pearson_correlation_dict(
        self,
        nrows=None,
        include_index=False,
        callback=None,
        extra_stats=False,
        min_shared=25,
        random_seed=0,
    ):
        """
        Calculates Pearson correlation coefficient between all pairs of columns in the DataFrame that
        support correlation. Works with numeric and datetime data. Call woodwork.utils.get_valid_pearson_types to
        see which Logical Types are supported.

        Args:
            nrows (int): The number of rows to sample for when determining correlation.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for correlation calculations.
                If False, the index column will not have the Pearson correlation calculated for it.
                Defaults to False.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:
                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            extra_stats (bool):  If True, additional column "shared_rows"
                recording the number of shared non-null rows for a column
                pair will be included with the dataframe. Defaults to False.
            min_shared (int): The number of shared non-null rows needed to
                calculate. Less rows than this will be considered too sparse
                to measure accurately and will return a NaN value. Must be
                non-negative. Defaults to 25.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        Returns:
            list(dict): A list containing dictionaries that have keys `column_1`,
            `column_2`, and `pearson` that is sorted in decending order by correlation coefficient.
            Correlation coefficient values are between -1 and 1.
        """
        return _get_dependence_dict(
            dataframe=self._dataframe,
            measures=["pearson"],
            nrows=nrows,
            include_index=include_index,
            callback=callback,
            extra_stats=extra_stats,
            min_shared=min_shared,
            random_seed=random_seed,
        )

    def pearson_correlation(
        self,
        nrows=None,
        include_index=False,
        callback=None,
        extra_stats=False,
        min_shared=25,
        random_seed=0,
    ):
        """Calculates Pearson correlation coefficient between all pairs of columns in the DataFrame that
        support correlation. Works with numeric and datetime data. Call woodwork.utils.get_valid_pearson_types to
        see which Logical Types are supported.

        Args:
            nrows (int): The number of rows to sample for when determining correlation.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for correlation calculations.
                If False, the index column will not have the Pearson correlation calculated for it.
                Defaults to False.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:
                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            extra_stats (bool):  If True, additional column "shared_rows"
                recording the number of shared non-null rows for a column
                pair will be included with the dataframe. Defaults to False.
            min_shared (int): The number of shared non-null rows needed to
                calculate. Less rows than this will be considered too sparse
                to measure accurately and will return a NaN value. Must be
                non-negative. Defaults to 25.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        Returns:
            pd.DataFrame: A DataFrame containing Pearson correlation coefficients with columns `column_1`,
            `column_2`, and `pearson` that is sorted in decending order by correlation value.
            Pearson values are between -1 and 1, with 0 meaning no correlation.
        """
        pearson_dict = self.pearson_correlation_dict(
            nrows=nrows,
            include_index=include_index,
            callback=callback,
            extra_stats=extra_stats,
            min_shared=min_shared,
            random_seed=random_seed,
        )
        return pd.DataFrame(pearson_dict)

    @_check_table_schema
    def dependence_dict(
        self,
        measures="all",
        num_bins=10,
        nrows=None,
        include_index=False,
        callback=None,
        extra_stats=False,
        min_shared=25,
        random_seed=0,
    ):
        """Calculates dependence measures between all pairs of columns in the DataFrame that
        support measuring dependence. Supports boolean, categorical, datetime, and numeric data.
        Call woodwork.utils.get_valid_mi_types and woodwork.utils.get_valid_pearson_types
        for complete lists of supported Logical Types.

        Args:
            dataframe (pd.DataFrame): Data containing Woodwork typing information
                from which to calculate dependence.
            measures (list or str): Which dependence measures to calculate.
                A list of measures can be provided to calculate multiple
                measures at once.  Valid measure strings:

                - "pearson": calculates the Pearson correlation coefficient
                - "mutual_info": calculates the mutual information between columns
                - "max":  max(abs(pearson), mutual) for each pair of columns
                - "all": includes columns for "pearson", "mutual_info", and "max"
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical. Defaults to 10. Pearson
                calculation does not use binning.
            nrows (int): The number of rows to sample for when determining dependence.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for measuring dependence.
                If False, the index column will not be considered. Defaults to False.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            extra_stats (bool):  If True, additional column "shared_rows"
                recording the number of shared non-null rows for a column
                pair will be included with the dataframe. Defaults to False. If
                the "max" measure is being used, a "measure_used" column will be
                added that records whether Pearson or mutual information was the
                maximum dependence for a particular row.
            min_shared (int): The number of shared non-null rows needed to
                calculate. Less rows than this will be considered too sparse
                to measure accurately and will return a NaN value. Must be
                non-negative. Defaults to 25.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        Returns:
            list(dict): A list containing dictionaries that have keys `column_1`,
            `column_2`, and keys for the specified dependence measures. The list is
            sorted in decending order by the first specified measure.
            Dependence information values are between 0 (no dependence) and 1
            (perfect dependency). For Pearson, values range from -1 to 1 but 0 is
            still no dependence.
        """
        return _get_dependence_dict(
            dataframe=self._dataframe,
            measures=measures,
            num_bins=num_bins,
            nrows=nrows,
            include_index=include_index,
            callback=callback,
            extra_stats=extra_stats,
            min_shared=min_shared,
            random_seed=random_seed,
        )

    def dependence(
        self,
        measures="all",
        num_bins=10,
        nrows=None,
        include_index=False,
        callback=None,
        extra_stats=False,
        min_shared=25,
        random_seed=0,
    ):
        """Calculates dependence measures between all pairs of columns in the DataFrame that
        support measuring dependence. Supports boolean, categorical, datetime, and numeric data.
        Call woodwork.utils.get_valid_mi_types and woodwork.utils.get_valid_pearson_types
        for complete lists of supported Logical Types.

        Args:
            dataframe (pd.DataFrame): Data containing Woodwork typing information
                from which to calculate dependence.
            measures (list or str): Which dependence measures to calculate.
                A list of measures can be provided to calculate multiple
                measures at once.  Valid measure strings:

                - "pearson": calculates the Pearson correlation coefficient
                - "mutual_info": calculates the mutual information between columns
                - "max":  max(abs(pearson), mutual) for each pair of columns
                - "all": includes columns for "pearson", "mutual_info", and "max"
            num_bins (int): Determines number of bins to use for converting
                numeric features into categorical. Defaults to 10. Pearson
                calculation does not use binning.
            nrows (int): The number of rows to sample for when determining dependence.
                If specified, samples the desired number of rows from the data.
                Defaults to using all rows.
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for measuring dependence.
                If False, the index column will not be considered. Defaults to False.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            extra_stats (bool):  If True, additional column "shared_rows"
                recording the number of shared non-null rows for a column
                pair will be included with the dataframe. Defaults to False. If
                the "max" measure is being used, a "measure_used" column will be
                added that records whether Pearson or mutual information was the
                maximum dependence for a particular row.
            min_shared (int): The number of shared non-null rows needed to
                calculate. Less rows than this will be considered too sparse
                to measure accurately and will return a NaN value. Must be
                non-negative. Defaults to 25.
            random_seed (int): Seed for the random number generator. Defaults to 0.
        Returns:
            pd.DataFrame: A DataFrame with the columns `column_1`,
            `column_2`, and keys for the specified dependence measures. The rows
            are sorted in decending order by the first specified measure.
            Dependence information values are between 0 (no dependence) and 1
            (perfect dependency). For Pearson, values range from -1 to 1 but 0 is
            still no dependence.  Additional columns will be included if the
            `extra_stats` is True.
        """
        dep_dict = _get_dependence_dict(
            dataframe=self._dataframe,
            measures=measures,
            num_bins=num_bins,
            nrows=nrows,
            include_index=include_index,
            callback=callback,
            extra_stats=extra_stats,
            min_shared=min_shared,
            random_seed=random_seed,
        )
        return pd.DataFrame(dep_dict)

    def get_valid_mi_columns(self, include_index=False):
        """Retrieves a list of columns from the DataFrame with valid Logical Types that support mutual
        information. Call woodwork.utils.get_valid_mi_types to see which Logical Types support mutual information.

        Args:
            include_index (bool): If True, the column specified as the index will be
                included as long as its LogicalType is valid for mutual information calculations.
                If False, the index column will not have mutual information calculated for it.
                Defaults to False.

        Returns:
            list: A list of column names that have valid Logical Types that support mutual information.
        """
        return _get_valid_mi_columns(self._dataframe, include_index)

    @_check_table_schema
    def describe_dict(
        self,
        include: Sequence[Union[str, LogicalType]] = None,
        callback: Callable[[int, int, int, str, float], Any] = None,
        results_callback: Callable[[pd.DataFrame, pd.Series], Any] = None,
        extra_stats: bool = False,
        bins: int = 10,
        top_x: int = 10,
        recent_x: int = 10,
    ) -> Dict[str, dict]:
        """Calculates statistics for data contained in the DataFrame.

        Args:
            include (list[str or LogicalType], optional): filter for what columns to include in the
                statistics returned. Can be a list of column names, semantic tags, logical types, or a list
                combining any of the three. It follows the most broad specification. Favors logical types
                then semantic tag then column name. If no matching columns are found, an empty DataFrame
                will be returned.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            results_callback (callable, optional): function to be called with intermediate results. Has the following parameters:

                - results_so_far (pd.DataFrame): the full dataframe calculated so far
                - most_recent_calculation (pd.Series): the calculations for the most recent column

            extra_stats (bool): If True, will calculate a histogram for numeric columns, top values
                for categorical columns and value counts for the most recent values in datetime columns. Will also
                calculate value counts within the range of values present for integer columns if the range of
                values present is less than or equal to than the number of bins used to compute the histogram.
                Output can be controlled by bins, top_x and recent_x parameters.
            bins (int): Number of bins to use when calculating histogram for numeric columns. Defaults to 10.
                Will be ignored unless extra_stats=True.
            top_x (int): Number of items to return when getting the most frequently occurring values for categorical
                columns. Defaults to 10. Will be ignored unless extra_stats=True.
            recent_x (int): Number of values to return when calculating value counts for the most recent dates in
                datetime columns. Defaults to 10. Will be ignored unless extra_stats=True.

        Returns:
            Dict[str -> dict]: A dictionary with a key for each column in the data or for each column
            matching the logical types, semantic tags or column names specified in ``include``, paired
            with a value containing a dictionary containing relevant statistics for that column.
        """
        return _get_describe_dict(
            self._dataframe,
            include=include,
            callback=callback,
            results_callback=results_callback,
            extra_stats=extra_stats,
            bins=bins,
            top_x=top_x,
            recent_x=recent_x,
        )

    def describe(
        self,
        include: Sequence[Union[str, LogicalType]] = None,
        callback: Callable[[int, int, int, str, float], Any] = None,
        results_callback: Callable[[pd.DataFrame, pd.Series], Any] = None,
    ) -> pd.DataFrame:
        """Calculates statistics for data contained in the DataFrame.

        Args:
            include (list[str or LogicalType], optional): filter for what columns to include in the
                statistics returned. Can be a list of column names, semantic tags, logical types, or a list
                combining any of the three. It follows the most broad specification. Favors logical types
                then semantic tag then column name. If no matching columns are found, an empty DataFrame
                will be returned.
            callback (callable, optional): Function to be called with incremental updates. Has the following parameters:

                - update (int): change in progress since last call
                - progress (int): the progress so far in the calculations
                - total (int): the total number of calculations to do
                - unit (str): unit of measurement for progress/total
                - time_elapsed (float): total time in seconds elapsed since start of call
            results_callback (callable, optional): function to be called with intermediate results. Has the following parameters:

                - results_so_far (pd.DataFrame): the full dataframe calculated so far
                - most_recent_calculation (pd.Series): the calculations for the most recent column

        Returns:
            pd.DataFrame: A Dataframe containing statistics for the data or the subset of the original
            DataFrame that contains the logical types, semantic tags, or column names specified
            in ``include``.
        """
        results = self.describe_dict(
            include=include, callback=callback, results_callback=results_callback
        )
        index_order = [
            "physical_type",
            "logical_type",
            "semantic_tags",
            "count",
            "nunique",
            "nan_count",
            "mean",
            "mode",
            "std",
            "min",
            "first_quartile",
            "second_quartile",
            "third_quartile",
            "max",
            "num_true",
            "num_false",
        ]
        return pd.DataFrame(results).reindex(index_order)

    @_check_table_schema
    def value_counts(self, ascending=False, top_n=10, dropna=False):
        """Returns a list of dictionaries with counts for the most frequent values in each column (only
            for columns with `category` as a standard tag).


        Args:
            ascending (bool): Defines whether each list of values should be sorted most frequent
                to least frequent value (False), or least frequent to most frequent value (True).
                Defaults to False.

            top_n (int): the number of top values to retrieve. Defaults to 10.

            dropna (bool): determines whether to remove NaN values when finding frequency. Defaults
                to False.

        Returns:
            list(dict): a list of dictionaries for each categorical column with keys `count`
            and `value`.
        """
        return _get_value_counts(self._dataframe, ascending, top_n, dropna)

    @_check_table_schema
    def infer_temporal_frequencies(self, temporal_columns=None, debug=False):
        """Infers the observation frequency (daily, biweekly, yearly, etc) of each temporal column
            in the DataFrame. Temporal columns are ones with the logical type Datetime or Timedelta.
            Not supported for Dask and Spark DataFrames.

        Args:
            temporal_columns (list[str], optional): Columns for which frequencies should be inferred. Must be columns
                that are present in the DataFrame and are temporal in nature. Defaults to None. If not
                specified, all temporal columns will have their frequencies inferred.
            debug (boolean): A flag used to determine if more information should be returned for each temporal column if
                no uniform frequency was found.

        Returns:
            (dict): A dictionary where each key is a temporal column from the DataFrame, and the
                value is its observation frequency represented as a pandas offset alias string (D, M, Y, etc.)
                or None if no uniform frequency was present in the data.

        Note:
            The pandas util ``pd.infer_freq``, which is used in this method, has the following behaviors:
                - If even one row in a column does not follow the frequency seen in the remaining rows,
                    no frequency will be inferred. Example of otherwise daily data that skips one day:
                    ``['2011-01-03', '2011-01-04', '2011-01-05', '2011-01-07']``.
                - If any NaNs are present in the data, no frequency will be inferred.
                - Pandas will use the largest offset alias available to it, so ``W`` will be inferred for weekly data instead of ``7D``.
                    The list of available offset aliases, which include aliases such as ``B`` for business day or ``N`` for nanosecond,
                    can be found at https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
                - Offset aliases can be combined to create something like ``2d1H``, which could also be expressed as '49H'.
                    Pandas' frequency inference will return the lower common alias, ``49H``, in situations when it'd otherwise
                    need to combine aliases.
                - Offset strings can contain more information than just the offset alias. For example, a date range
                    ``pd.date_range(start="2020-01-01", freq="w", periods=10)`` will be inferred to have frequency ``W-SUN``.
                    That string is an offset alias with an anchoring suffix that indicates that the data is not only
                    observed at a weekly frequency, but that all the dates are on Sundays. More anchored offsets
                    can be seen here: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#anchored-offsets
                - Some frequencies that can be defined for a ``pd.date_range`` cannot then be re-inferred by pandas' ``pd.infer_freq``.
                    One example of this can be seen when using the business day offset alias ``B``
                    ``pd.date_range(start="2020-01-01", freq="4b", periods=10)``, which is a valid ``freq``
                    parameter when building the date range, but is not then inferrable.

            The algorithm used to infer frequency on noisy data can configured (see https://woodwork.alteryx.com/en/stable/guides/setting_config_options.html#Available-Config-Settings)

        """
        return _infer_temporal_frequencies(
            self._dataframe, temporal_columns=temporal_columns, debug=debug
        )

    @_check_table_schema
    def validate_logical_types(self, return_invalid_values=False):
        """Validates the dataframe based on the logical types.
        If a column's dtype does not match its logical type's required dtype,
        will raise a TypeValidationError even when return_invalid_indices is True.

        Args:
            return_invalid_values (bool): Whether or not to return invalid data values

        Returns:
            DataFrame: If return_invalid_values is True, returns invalid data values.
        """
        invalid_values = []
        for column in self.columns:
            series = self.ww[column]
            values = series.ww.validate_logical_type(
                return_invalid_values=return_invalid_values
            )
            if values is not None:
                invalid_values.append(values)

        if return_invalid_values:
            concat = pd.concat
            if _is_dask_dataframe(self._dataframe):
                concat = dd.concat
            if _is_spark_dataframe(self._dataframe):
                concat = ps.concat

            return concat(invalid_values, axis=1)


def _validate_accessor_params(
    dataframe, index, time_index, logical_types, schema, use_standard_tags
) -> None:
    _check_unique_column_names(dataframe)
    if use_standard_tags is not None:
        _check_use_standard_tags(use_standard_tags)
    if schema is not None:
        _check_partial_schema(dataframe, schema)
        if index is None:
            # if no index was passed in as a parameter we need to validate the existing index
            index = schema.index
    if index is not None:
        _check_index(dataframe, index)
    if logical_types:
        _check_logical_types(dataframe.columns, logical_types)
    if time_index is not None:
        datetime_format = None
        logical_type = None
        if logical_types is not None and time_index in logical_types:
            logical_type = logical_types[time_index]
            if type(logical_types[time_index]) == Datetime:
                datetime_format = logical_types[time_index].datetime_format

        _check_time_index(
            dataframe,
            time_index,
            datetime_format=datetime_format,
            logical_type=logical_type,
        )


def _check_unique_column_names(dataframe):
    if not dataframe.columns.is_unique:
        raise IndexError("Dataframe cannot contain duplicate columns names")


def _check_index(dataframe, index):
    if index not in dataframe.columns:
        # User specifies an index that is not in the dataframe
        raise ColumnNotPresentError(
            f"Specified index column `{index}` not found in dataframe"
        )
    if index is not None and isinstance(dataframe, pd.DataFrame):
        # User specifies a dataframe index that is not unique or contains null values
        # Does not check Dask dataframes to avoid pulling data into memory and Dask does not support is_unique
        if not dataframe[index].is_unique:
            raise IndexError("Index column must be unique")

        if dataframe[index].isnull().any():
            raise IndexError("Index contains null values")


def _check_time_index(dataframe, time_index, datetime_format=None, logical_type=None):
    if time_index not in dataframe.columns:
        raise ColumnNotPresentError(
            f"Specified time index column `{time_index}` not found in dataframe"
        )
    if not (
        _is_numeric_series(dataframe[time_index], logical_type)
        or col_is_datetime(dataframe[time_index], datetime_format=datetime_format)
    ):
        raise TypeError("Time index column must contain datetime or numeric values")


def _check_logical_types(dataframe_columns, logical_types):
    if not isinstance(logical_types, dict):
        raise TypeError("logical_types must be a dictionary")
    cols_not_found = set(logical_types.keys()).difference(set(dataframe_columns))
    if cols_not_found:
        raise ColumnNotPresentError(
            "logical_types contains columns that are not present in "
            f"dataframe: {sorted(list(cols_not_found))}"
        )


def _check_schema(dataframe, schema):
    if not isinstance(schema, TableSchema):
        raise TypeError("Provided schema must be a Woodwork.TableSchema object.")
    invalid_schema_message = get_invalid_schema_message(dataframe, schema)
    if invalid_schema_message:
        raise ValueError(
            f"Woodwork typing information is not valid for this DataFrame: {invalid_schema_message}"
        )


def _check_partial_schema(dataframe, schema: TableSchema) -> None:
    if not isinstance(schema, TableSchema):
        raise TypeError("Provided schema must be a Woodwork.TableSchema object.")
    dataframe_cols = set(dataframe.columns)
    schema_cols = set(schema.columns.keys())
    schema_cols_not_in_df = schema_cols - dataframe_cols
    if schema_cols_not_in_df:
        raise ColumnNotPresentError(
            f"The following columns in the typing information were missing from the DataFrame: "
            f"{schema_cols_not_in_df}"
        )


def _check_use_standard_tags(use_standard_tags):
    if not isinstance(use_standard_tags, (bool, dict)):
        raise TypeError("use_standard_tags must be a dictionary or a boolean")


def _infer_missing_logical_types(
    dataframe: AnyDataFrame,
    force_logical_types: Dict[ColumnName, Union[str, LogicalType]] = None,
    existing_logical_types: Dict[ColumnName, Union[str, LogicalType]] = None,
):
    """Performs type inference and updates underlying data"""
    force_logical_types = force_logical_types or {}
    existing_logical_types = existing_logical_types or {}
    parsed_logical_types = {}
    for name in dataframe.columns:
        series = dataframe[name]
        logical_type = (
            force_logical_types.get(name)
            if name in force_logical_types
            else existing_logical_types.get(name)
        )
        parsed_logical_types[name] = _get_column_logical_type(
            series, logical_type, name
        )
        updated_series = parsed_logical_types[name].transform(series)
        if updated_series is not series:
            dataframe[name] = updated_series
    return parsed_logical_types


def _merge_use_standard_tags(
    existing_use_standard_tags: UseStandardTagsDict = None,
    use_standard_tags: Union[bool, UseStandardTagsDict] = None,
    column_names: Iterable[ColumnName] = None,
    default_use_standard_tag: bool = True,
) -> UseStandardTagsDict:
    """Combines existing and kwarg use_standard_tags and returns a UseStandardTagsDict with all column names
    Priority when merging:
    1. use_standard tags
    2. existing_use_standard_tags
    3. default_use_standard_tag
    """
    column_names = column_names or []
    if isinstance(use_standard_tags, bool):
        use_standard_tags = {col_name: use_standard_tags for col_name in column_names}
    else:
        use_standard_tags = {
            **{col_name: default_use_standard_tag for col_name in column_names},
            **(existing_use_standard_tags or {}),
            **(use_standard_tags or {}),
        }
    return use_standard_tags


@pd.api.extensions.register_dataframe_accessor("ww")
class PandasTableAccessor(WoodworkTableAccessor):
    pass


if dd:

    @dd.extensions.register_dataframe_accessor("ww")
    class DaskTableAccessor(WoodworkTableAccessor):
        pass


if ps:
    from pyspark.pandas.extensions import register_dataframe_accessor

    @register_dataframe_accessor("ww")
    class SparkTableAccessor(WoodworkTableAccessor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if not ps.get_option("compute.ops_on_diff_frames"):
                ps.set_option("compute.ops_on_diff_frames", True)



# flake8: noqa
import pytest
from .table_utils import (
    _check_close,
    check_empty_box_plot_dict,
    dep_between_cols,
    is_property,
    is_public_method,
    to_pandas,
    validate_subset_schema,
)


def pd_to_dask(series):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(series, npartitions=2)


def pd_to_spark(series):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(convert_tuples_to_lists(series))


def convert_tuples_to_lists(series):
    def apply_func(value):
        if type(value) is tuple:
            return list(value)
        return value

    return series.apply(apply_func)



from .parquet_serializer import ParquetSerializer
from .csv_serializer import CSVSerializer
from .feather_serializer import FeatherSerializer
from .arrow_serializer import ArrowSerializer
from .orc_serializer import OrcSerializer
from .pickle_serializer import PickleSerializer

from .utils import get_serializer



import json
import os
import warnings
from pathlib import Path

import pandas as pd

from woodwork.accessor_utils import _is_dask_dataframe, _is_spark_dataframe
from woodwork.exceptions import ParametersIgnoredWarning, WoodworkFileExistsError
from woodwork.serializers.serializer_base import (
    PYARROW_IMPORT_ERROR_MESSAGE,
    Serializer,
    clean_latlong,
)
from woodwork.utils import import_or_raise


class ParquetSerializer(Serializer):
    """Serialize a Woodwork table to disk as a parquet file."""

    format = "parquet"

    def __init__(self, path, filename, data_subdirectory, typing_info_filename):
        import_or_raise("pyarrow", PYARROW_IMPORT_ERROR_MESSAGE)
        super().__init__(path, filename, data_subdirectory, typing_info_filename)
        if typing_info_filename and typing_info_filename != "woodwork_typing_info.json":
            warnings.warn(
                "Typing info filename has been ignored. Typing information will be stored in parquet file header.",
                ParametersIgnoredWarning,
            )
        self.typing_info_filename = None

    def serialize(self, dataframe, profile_name, **kwargs):
        if self.filename is not None and _is_dask_dataframe(dataframe):
            raise ValueError(
                "Writing a Dask dataframe to parquet with a filename specified is not supported"
            )
        if self.filename is not None and _is_spark_dataframe(dataframe):
            raise ValueError(
                "Writing a Spark dataframe to parquet with a filename specified is not supported"
            )
        self.kwargs["engine"] = "pyarrow"
        return super().serialize(dataframe, profile_name, **kwargs)

    def save_to_local_path(self):
        """Serialize data and typing information to a local directory. Overrides method on base class
        due to different serialiation flow"""
        if self.data_subdirectory:
            location = os.path.join(self.write_path, self.data_subdirectory)
            os.makedirs(location, exist_ok=True)
        else:
            os.makedirs(self.write_path, exist_ok=True)
        self._create_pyarrow_table()
        self._generate_parquet_metadata()
        self._save_parquet_table_to_disk()

    def _create_pyarrow_table(self):
        """Create a pyarrow table for pandas. This table will get updated to included
        Woodwork typing info before saving. Skip for Dask/Spark because for those formats
        typing information has to be added after files are saved to disk."""
        import pyarrow as pa

        if isinstance(self.dataframe, pd.DataFrame):
            dataframe = clean_latlong(self.dataframe)
            self.table = pa.Table.from_pandas(dataframe)

    def _generate_parquet_metadata(self):
        """Generate metadata for the parquet file header. For pandas this includes additional
        information needed by pandas. For Dask/Spark, this includes only the Woodwork typing info."""
        loading_info = {
            "location": self.location,
            "type": self.format,
            "params": self.kwargs,
        }
        self.typing_info["loading_info"].update(loading_info)
        # For Dask and Spark we only get the WW metadata because we haven't created
        # the pyarrow table yet, but for pandas we combine the existing parquet
        # metadata with the WW metadata.
        if _is_dask_dataframe(self.dataframe) or _is_spark_dataframe(self.dataframe):
            metadata = {
                "ww_meta".encode(): json.dumps(self.typing_info).encode(),
            }
        else:
            table_metadata = self.table.schema.metadata
            metadata = {
                "ww_meta".encode(): json.dumps(self.typing_info).encode(),
                **table_metadata,
            }
        self.metadata = metadata

    def _save_parquet_table_to_disk(self):
        """Writes data to disk with the updated metadata including WW typing info."""
        import pyarrow.parquet as pq

        if _is_dask_dataframe(self.dataframe):
            path, dataframe = self._setup_for_dask_and_spark()
            dataframe.to_parquet(path, custom_metadata=self.metadata)
        elif _is_spark_dataframe(self.dataframe):
            path, dataframe = self._setup_for_dask_and_spark()
            dataframe.to_parquet(path)
            files = os.listdir(path)

            # Update first parquet file to save WW metadata
            parquet_files = sorted([f for f in files if Path(f).suffix == ".parquet"])
            update_file = os.path.join(path, parquet_files[0])
            table = pq.read_table(update_file)
            table_metadata = table.schema.metadata
            combined_meta = {
                **self.metadata,
                **table_metadata,
            }
            table = table.replace_schema_metadata(combined_meta)
            pq.write_table(table, update_file)

            # Remove checksum files which prevent deserialization if present due to updated parquet header
            crc_files = [f for f in files if Path(f).suffix == ".crc"]
            for file in crc_files:
                os.remove(os.path.join(path, file))
        else:
            file = self._get_filename()
            self.table = self.table.replace_schema_metadata(self.metadata)
            pq.write_table(self.table, file)

    def _setup_for_dask_and_spark(self):
        """Perform additional path setup required for Dask/Spark. Since Dask/Spark deserialize to
        directories only, the `_get_filename` method does not work like it does for pandas."""
        path = self.path
        if self.data_subdirectory is not None:
            path = os.path.join(path, self.data_subdirectory)
        if any([Path(f).suffix == ".parquet" for f in os.listdir(path)]):
            message = f"Data file already exists at '{path}'. "
            message += "Please remove or use a different directory."
            raise WoodworkFileExistsError(message)
        return path, clean_latlong(self.dataframe)



from ._bin_numeric_cols_into_categories import _bin_numeric_cols_into_categories
from ._get_box_plot_info_for_column import _get_box_plot_info_for_column
from ._get_dependence_dict import _get_dependence_dict
from ._get_describe_dict import _get_describe_dict
from ._get_histogram_values import _get_histogram_values
from ._get_mode import _get_mode
from ._get_numeric_value_counts_in_range import _get_numeric_value_counts_in_range
from ._get_recent_value_counts import _get_recent_value_counts
from ._get_top_values_categorical import _get_top_values_categorical
from ._get_valid_mi_columns import _get_valid_mi_columns
from ._get_value_counts import _get_value_counts
from ._infer_temporal_frequencies import _infer_temporal_frequencies
from .frequency_inference._infer_frequency import infer_frequency



import json
import os
import tarfile
import tempfile
from pathlib import Path

from woodwork.deserializers.deserializer_base import (
    PYARROW_IMPORT_ERROR_MESSAGE_DESERIALIZE,
    Deserializer,
    _check_schema_version,
)
from woodwork.s3_utils import get_transport_params, use_smartopen
from woodwork.utils import _is_s3, _is_url, import_or_raise


class ParquetDeserializer(Deserializer):
    """Deserialize Woodwork table from serialized data in parquet format."""

    format = "parquet"

    def deserialize(self, profile_name, validate):
        import_or_raise("pyarrow", PYARROW_IMPORT_ERROR_MESSAGE_DESERIALIZE)
        if _is_url(self.path) or _is_s3(self.path):
            dataframe = self.read_from_s3(profile_name)
        else:
            if self.data_subdirectory:
                self.path = os.path.join(self.path, self.data_subdirectory)
            self.read_path = self.path
            if self.filename:
                self.read_path = os.path.join(self.path, self.filename)

            dataframe = self.read_from_local_path()
        dataframe.ww.init(**self.ww_init_dict, validate=validate)
        return dataframe

    def configure_deserializer(self):
        import pyarrow as pa

        self._set_metadata_path()
        file_metadata = pa.parquet.read_metadata(self.metadata_path)
        self.typing_info = json.loads(file_metadata.metadata[b"ww_meta"])
        _check_schema_version(self.typing_info["schema_version"])
        loading_info = self.typing_info["loading_info"]
        self.kwargs = loading_info.get("params", {})
        self._set_init_dict(loading_info)

    def read_from_local_path(self):
        self.configure_deserializer()
        lib = self._get_library()
        return lib.read_parquet(self.read_path, engine=self.kwargs["engine"])

    def read_from_s3(self, profile_name):
        with tempfile.TemporaryDirectory() as tmpdir:
            tar_filename = Path(self.path).name
            tar_filepath = os.path.join(tmpdir, tar_filename)
            transport_params = None

            if _is_s3(self.path):
                transport_params = get_transport_params(profile_name)

            use_smartopen(tar_filepath, self.path, transport_params)
            with tarfile.open(str(tar_filepath)) as tar:
                tar.extractall(path=tmpdir)

            self.read_path = os.path.join(tmpdir, self.data_subdirectory, self.filename)

            return self.read_from_local_path()

    def _set_metadata_path(self):
        # If we are reading a single pandas file, we get the metadata from the file.
        # If we are reading into Dask/Spark we need to get the metadata from the
        # first file that was serialized.
        self.metadata_path = self.read_path
        if os.path.isdir(self.read_path):
            files = os.listdir(self.read_path)
            if "part.0.parquet" in files:
                # Dask will serialize with "part.*.parquet" file names
                self.metadata_path = os.path.join(self.read_path, "part.0.parquet")
            elif any(["snappy.parquet" in f for f in files]):
                # Spark will serialize files with a unique hash but with the ".snappy.parquet" extension
                parquet_files = sorted(
                    [f for f in files if Path(f).suffix == ".parquet"]
                )
                self.metadata_path = os.path.join(self.read_path, parquet_files[0])



import numpy as np
import pandas as pd
import pytest

from woodwork.logical_types import (
    URL,
    Boolean,
    BooleanNullable,
    Categorical,
    Datetime,
    Double,
    EmailAddress,
    Integer,
    IntegerNullable,
    IPAddress,
    PersonFullName,
    PhoneNumber,
    Unknown,
)
from woodwork.tests.testing_utils import pd_to_dask, pd_to_spark
from woodwork.utils import import_or_none


@pytest.fixture(scope="session", autouse=True)
def spark_session():
    pyspark = import_or_none("pyspark.sql")

    if pyspark:
        spark = (
            pyspark.SparkSession.builder.master("local[*]")
            .config(
                "spark.driver.extraJavaOptions",
                "-Dio.netty.tryReflectionSetAccessible=True",
            )
            .config("spark.sql.shuffle.partitions", "2")
            .config("spark.driver.bindAddress", "127.0.0.1")
            .getOrCreate()
        )

        return spark


@pytest.fixture(params=["sample_df_pandas", "sample_df_dask", "sample_df_spark"])
def sample_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_df_pandas():
    return pd.DataFrame(
        {
            "id": range(4),
            "full_name": [
                "Mr. John Doe",
                "Doe, Mrs. Jane",
                "James Brown",
                "Ms. Paige Turner",
            ],
            "email": [
                "john.smith@example.com",
                np.nan,
                "team@featuretools.com",
                "junk@example.com",
            ],
            "phone_number": [
                "5555555555",
                "555-555-5555",
                "1-(555)-555-5555",
                "555-555-5555",
            ],
            "age": pd.Series([pd.NA, 33, 33, 57], dtype="Int64"),
            "signup_date": [pd.to_datetime("2020-09-01")] * 4,
            "is_registered": pd.Series([pd.NA, False, True, True], dtype="boolean"),
            "double": [0, 1, 0.0001, -5.4321],
            "double_with_nan": [np.nan, -123.45, 0.0001, 1],
            "integer": [-1, 4, 9, 25],
            "nullable_integer": pd.Series([pd.NA, -36, 49, 64], dtype="Int64"),
            "boolean": [True, False, False, True],
            "categorical": pd.Series(["a", "b", "c", "a"], dtype="category"),
            "datetime_with_NaT": [pd.to_datetime("2020-09-01")] * 3 + [pd.NaT],
            "url": [
                "https://github.com/alteryx/woodwork",
                "https://twitter.com/AlteryxOSS",
                np.nan,
                "http://google.com",
            ],
            "ip_address": [
                "172.16.254.1",
                "2001:0db8:85a3:0000:0000:8a2e:0370:7334",
                "1762:0:0:0:0:B03:1:AF18",
                np.nan,
            ],
        }
    )


@pytest.fixture()
def sample_df_dask(sample_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(sample_df_pandas, npartitions=2)


@pytest.fixture()
def sample_df_spark(sample_df_pandas):
    ps = pytest.importorskip(
        "pyspark.pandas", reason="Pyspark pandas not installed, skipping"
    )
    return ps.from_pandas(sample_df_pandas)


@pytest.fixture()
def sample_df_phone_numbers():
    return pd.DataFrame(
        {
            "phone_number": [
                "2002007865",
                "311-311-3156",
                "422.422.0461",
                "533 533 8122",
                "644●644●7865",
                "(755) 755 7109",
                "(866) 866-0182",
                "+1 (977) 977 1779",
                "+1(288)-288-7772",
                "+1 399 399 2933",
                "+1-400-400-1109",
                "+1.511.511.2289",
                "+16226229222",
                "17336330019",
                "1.844.744.1842",
                "1●955●855●9272",
                "1 266 966 2283",
                "+001 236 248 8482",
            ]
        }
    )


@pytest.fixture()
def sample_df_postal_code():
    return pd.DataFrame(
        {
            "postal_code": [
                "20020",
                "07865",
                "12345-6789",
                "11111-1111",
            ]
        }
    )


@pytest.fixture()
def datetime_freqs_df_pandas():
    return pd.DataFrame(
        {
            "2D_freq": pd.date_range(start="2020-01-01", end="2020-01-20", freq="2D"),
            "3M_freq": pd.date_range(start="2015-01-01", freq="3M", periods=10),
            "3B_no_freq": pd.date_range(start="2015-01-01", freq="3B", periods=10),
            "1d_skipped_one_freq": pd.date_range(
                start="2020-01-01", end="2020-01-11", freq="1D"
            ).drop("2020-01-04"),
            "3M_one_nan": list(
                pd.date_range(start="2015-01-01", freq="3M", periods=10).drop(
                    "2015-07-31"
                )
            )
            + [None],
            "2d_timedelta": pd.date_range(
                start="2020-01-01", end="2020-01-20", freq="2D"
            )
            - pd.Timestamp("2020-01-01"),
            "ints": range(10),
            "same_date": ["2015-01-01"] * 10,
        }
    )


@pytest.fixture(
    params=[
        "sample_unsorted_df_pandas",
        "sample_unsorted_df_dask",
        "sample_unsorted_df_spark",
    ]
)
def sample_unsorted_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_unsorted_df_pandas():
    return pd.DataFrame(
        {
            "id": [3, 1, 2, 0],
            "full_name": [
                "Mr. John Doe",
                "Doe, Mrs. Jane",
                "James Brown",
                "Ms. Paige Turner",
            ],
            "email": [
                "john.smith@example.com",
                np.nan,
                "team@featuretools.com",
                "junk@example.com",
            ],
            "phone_number": [
                "5555555555",
                "555-555-5555",
                "1-(555)-555-5555",
                "555-555-5555",
            ],
            "age": [33, 25, 33, 57],
            "signup_date": pd.to_datetime(
                ["2020-09-01", "2020-08-01", "2020-08-02", "2020-09-01"]
            ),
            "is_registered": [True, False, True, True],
        }
    )


@pytest.fixture()
def sample_unsorted_df_dask(sample_unsorted_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(sample_unsorted_df_pandas, npartitions=2)


@pytest.fixture()
def sample_unsorted_df_spark(sample_unsorted_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(sample_unsorted_df_pandas)


@pytest.fixture(
    params=["sample_series_pandas", "sample_series_dask", "sample_series_spark"]
)
def sample_series(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_series_pandas():
    return pd.Series(
        ["a", "b", "c"] + 10 * ["a", "a", "a"], name="sample_series"
    ).astype("category")


@pytest.fixture()
def sample_series_dask(sample_series_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(sample_series_pandas, npartitions=2)


@pytest.fixture()
def sample_series_spark(sample_series_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(sample_series_pandas.astype("string"))


@pytest.fixture(
    params=[
        "sample_datetime_series_pandas",
        "sample_datetime_series_dask",
        "sample_datetime_series_spark",
    ]
)
def sample_datetime_series(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_datetime_series_pandas():
    return pd.Series(
        [pd.to_datetime("2020-09-01")] * 4, name="sample_datetime_series"
    ).astype("object")


@pytest.fixture()
def sample_datetime_series_dask(sample_datetime_series_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(sample_datetime_series_pandas, npartitions=2)


@pytest.fixture()
def sample_datetime_series_spark(sample_datetime_series_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(sample_datetime_series_pandas.astype(str))


@pytest.fixture()
def ordinal_transform_series_pandas():
    return pd.Series([1, 2, 3], dtype="int64")


@pytest.fixture()
def ordinal_transform_series_dask(ordinal_transform_series_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(ordinal_transform_series_pandas, npartitions=1)


@pytest.fixture()
def ordinal_transform_series_spark(ordinal_transform_series_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(ordinal_transform_series_pandas)


@pytest.fixture()
def time_index_df_pandas():
    return pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "times": ["2019-01-01", "2019-01-02", "2019-01-03", pd.NA],
            "ints": [1, 2, 3, 4],
            "strs": ["1", "2", "3", "4"],
            "letters": ["a", "b", "c", "d"],
            "bools": [True, False, False, True],
        }
    )


@pytest.fixture()
def time_index_df_dask(time_index_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(time_index_df_pandas, npartitions=2)


@pytest.fixture()
def time_index_df_spark(time_index_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(time_index_df_pandas)


@pytest.fixture(
    params=["time_index_df_pandas", "time_index_df_dask", "time_index_df_spark"]
)
def time_index_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def numeric_time_index_df_pandas():
    return pd.DataFrame(
        {
            "floats": pd.Series([1, 2, 3, 4], dtype="float"),
            "ints": pd.Series([1, -2, 3, 4], dtype="int64"),
            "with_null": pd.Series([1, pd.NA, 3, 4], dtype="Int64"),
        }
    )


@pytest.fixture()
def numeric_time_index_df_dask(numeric_time_index_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(numeric_time_index_df_pandas, npartitions=2)


@pytest.fixture()
def numeric_time_index_df_spark(numeric_time_index_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(numeric_time_index_df_pandas)


@pytest.fixture(
    params=[
        "numeric_time_index_df_pandas",
        "numeric_time_index_df_dask",
        "numeric_time_index_df_spark",
    ]
)
def numeric_time_index_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def describe_df_pandas():
    index_data = [0, 1, 2, 3, 4, 5, 6, 7]
    boolean_data = [True, False, True, True, False, True, False, None]
    category_data = ["red", "blue", "red", np.nan, "red", "blue", "red", "yellow"]
    datetime_data = pd.to_datetime(
        [
            "2020-01-01",
            "2020-02-01",
            "2020-01-01 08:00",
            "2020-02-02 16:00",
            "2020-02-02 18:00",
            pd.NaT,
            "2020-02-01",
            "2020-01-02",
        ]
    )
    formatted_datetime_data = pd.Series(
        [
            "2020~01~01",
            "2020~02~01",
            "2020~03~01",
            "2020~02~02",
            "2020~03~02",
            pd.NaT,
            "2020~02~01",
            "2020~01~02",
        ]
    )
    numeric_data = pd.Series([10, 20, 17, 32, np.nan, 1, 56, 10])
    natural_language_data = [
        "This is a natural language sentence",
        "Duplicate sentence.",
        "This line has numbers in it 000123.",
        "How about some symbols?!",
        "This entry contains two sentences. Second sentence.",
        "Duplicate sentence.",
        np.nan,
        "I am the last line",
    ]
    latlong_data = [
        (0, 0),
        (1, 1),
        (2, 2),
        (3, 3),
        (0, 0),
        (np.nan, np.nan),
        (np.nan, 6),
        np.nan,
    ]
    timedelta_data = datetime_data - pd.Timestamp("2020-01-01")
    unknown_data = [
        "unknown1",
        "unknown5",
        "unknown3",
        np.nan,
        "unknown3",
        np.nan,
        "unknown5",
        "unknown5",
    ]

    return pd.DataFrame(
        {
            "index_col": index_data,
            "boolean_col": boolean_data,
            "category_col": category_data,
            "datetime_col": datetime_data,
            "formatted_datetime_col": formatted_datetime_data,
            "numeric_col": numeric_data,
            "natural_language_col": natural_language_data,
            "timedelta_col": timedelta_data,
            "latlong_col": latlong_data,
            "unknown_col": unknown_data,
        }
    )


@pytest.fixture()
def describe_df_dask(describe_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(describe_df_pandas, npartitions=2)


@pytest.fixture()
def describe_df_spark(describe_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(
        describe_df_pandas.applymap(
            lambda tup: [None if pd.isnull(elt) else elt for elt in tup]
            if isinstance(tup, tuple)
            else tup
        ).drop(columns="timedelta_col")
    )


@pytest.fixture(params=["describe_df_pandas", "describe_df_dask", "describe_df_spark"])
def describe_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def df_same_mi_pandas():
    return pd.DataFrame(
        {
            "ints": pd.Series([2, pd.NA, 5, 2], dtype="Int64"),
            "floats": pd.Series([1, None, 100, 1]),
            "nans": pd.Series([None, None, None, None]),
            "nat_lang": pd.Series(
                [
                    "this is a very long sentence inferred as a string",
                    None,
                    "test",
                    "test",
                ]
            ),
        }
    )


@pytest.fixture()
def df_same_mi_dask(df_same_mi_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(df_same_mi_pandas, npartitions=2)


@pytest.fixture()
def df_same_mi_spark(df_same_mi_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    df_same_mi_pandas["ints"] = df_same_mi_pandas["ints"].astype("float")
    df_same_mi_pandas["nans"] = df_same_mi_pandas["nans"].astype("float")
    return ps.from_pandas(df_same_mi_pandas)


@pytest.fixture(params=["df_same_mi_pandas", "df_same_mi_dask", "df_same_mi_spark"])
def df_same_mi(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def df_mi_pandas():
    df = pd.DataFrame(
        {
            "ints": pd.Series([1, 2, 1]),
            "ints2": pd.Series([2, 2, 2]),
            "bools": pd.Series([True, False, True]),
            "strs2": pd.Series(["bye", "hi", "bye"]),
            "strs": pd.Series(["hi", "hi", "hi"]),
            "dates": pd.Series(["2020-01-01", "2020-01-01", "1997-01-04"]),
        }
    )
    df = df.loc[df.index.repeat(4)].reset_index(drop=True)
    return df


@pytest.fixture()
def df_mi_dask(df_mi_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(df_mi_pandas, npartitions=1)


@pytest.fixture()
def df_mi_spark(df_mi_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(df_mi_pandas)


@pytest.fixture(params=["df_mi_pandas", "df_mi_dask", "df_mi_spark"])
def df_mi(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def df_mi_unique_pandas():
    df = pd.DataFrame(
        {
            "unique": pd.Series(["hi", "bye", "hello", "goodbye"]),
            "unique_with_one_nan": pd.Series(["hi", "bye", None, "goodbye"]),
            "unique_with_nans": pd.Series([1, None, None, 2]),
            "ints": pd.Series([1, 2, 1, 2]),
        }
    )
    df = df.loc[df.index.repeat(5)].reset_index(drop=True)
    return df


@pytest.fixture()
def df_mi_unique_dask(df_mi_unique_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(df_mi_unique_pandas, npartitions=1)


@pytest.fixture()
def df_mi_unique_spark(df_mi_unique_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(df_mi_unique_pandas)


@pytest.fixture(
    params=["df_mi_unique_pandas", "df_mi_unique_dask", "df_mi_unique_spark"]
)
def df_mi_unique(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def categorical_df_pandas():
    return pd.DataFrame(
        {
            "ints": pd.Series([1, 2, 3, 2]),
            "categories1": pd.Series([1, 100, 1, 100, 200, 200, 200, 200, 3, 100]),
            "bools": pd.Series([True, False, True, False]),
            "categories2": pd.Series(["test", "test", "test2", "test"]),
            "categories3": pd.Series(["test", "test", "test", np.nan]),
        }
    )


@pytest.fixture()
def categorical_df_dask(categorical_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(categorical_df_pandas, npartitions=2)


@pytest.fixture()
def categorical_df_spark(categorical_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(categorical_df_pandas)


@pytest.fixture(
    params=["categorical_df_pandas", "categorical_df_dask", "categorical_df_spark"]
)
def categorical_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def empty_df_pandas():
    return pd.DataFrame({})


@pytest.fixture()
def empty_df_dask(empty_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(empty_df_pandas, npartitions=2)


# Cannot have an empty Spark DataFrame
@pytest.fixture(params=["empty_df_pandas", "empty_df_dask"])
def empty_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def small_df_pandas():
    return pd.DataFrame(
        pd.Series(
            [pd.to_datetime("2020-09-01")] * 4, name="sample_datetime_series"
        ).astype("object")
    )


@pytest.fixture()
def small_df_dask(small_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(small_df_pandas, npartitions=1)


@pytest.fixture()
def small_df_spark(small_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    df = small_df_pandas.astype("datetime64[ns]")
    return ps.from_pandas(df)


@pytest.fixture(params=["small_df_pandas", "small_df_dask", "small_df_spark"])
def small_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def latlong_df_pandas():
    return pd.DataFrame(
        {
            "tuple_ints": pd.Series([(1, 2), (3, 4)]),
            "tuple_strings": pd.Series([("1", "2"), ("3", "4")]),
            "string_tuple": pd.Series(["(1, 2)", "(3, 4)"]),
            "bracketless_string_tuple": pd.Series(["1, 2", "3, 4"]),
            "list_strings": pd.Series([["1", "2"], ["3", "4"]]),
            "combo_tuple_types": pd.Series(["[1, 2]", "(3, 4)"]),
            "null_value": pd.Series([np.nan, (3, 4)]),
            "null_latitude": pd.Series([(np.nan, 2.0), (3.0, 4.0)]),
            "both_null": pd.Series([(np.nan, np.nan), (3.0, 4.0)]),
        }
    )


@pytest.fixture()
def latlong_df_dask(latlong_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(latlong_df_pandas, npartitions=2)


@pytest.fixture()
def latlong_df_spark(latlong_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(
        latlong_df_pandas.applymap(
            lambda tup: list(tup) if isinstance(tup, tuple) else tup
        )
    )


@pytest.fixture(params=["latlong_df_pandas", "latlong_df_dask", "latlong_df_spark"])
def latlong_df(request):
    return request.getfixturevalue(request.param)


# LatLong Fixtures for testing access to latlong values
@pytest.fixture
def pandas_latlongs():
    return [
        pd.Series([(1.0, 2.0), (3.0, 4.0)]),
        pd.Series([("1", "2"), ("3", "4")]),
        pd.Series([["1", "2"], ["3", "4"]]),
        pd.Series([(1, 2), (3, 4)]),
        pd.Series([[1, 2], [3, 4]]),
        pd.Series(["(1, 2)", "(3, 4)"]),
        pd.Series(["1, 2", "3, 4"]),
        pd.Series(["[1, 2]", "[3, 4]"]),
    ]


@pytest.fixture
def dask_latlongs(pandas_latlongs):
    return [pd_to_dask(series) for series in pandas_latlongs]


@pytest.fixture
def spark_latlongs(pandas_latlongs):
    return [
        pd_to_spark(
            series.apply(lambda tup: list(tup) if isinstance(tup, tuple) else tup)
        )
        for series in pandas_latlongs
    ]


@pytest.fixture(params=["pandas_latlongs", "dask_latlongs", "spark_latlongs"])
def latlongs(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def whitespace_df_pandas():
    return pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "comments": [
                "\nleading newline",
                "trailing newline\n",
                "\n",
                "newline in \n the middle",
                "    leading whitespace",
                "trailing whitespace ",
            ],
        }
    )


@pytest.fixture()
def whitespace_df_dask(whitespace_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(whitespace_df_pandas, npartitions=2)


@pytest.fixture()
def whitespace_df_spark(whitespace_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(whitespace_df_pandas)


@pytest.fixture(
    params=["whitespace_df_pandas", "whitespace_df_dask", "whitespace_df_spark"]
)
def whitespace_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def falsy_names_df_pandas():
    return pd.DataFrame(
        {
            0: ["a", "b", "c"],
            "": [1, 2, 3],
        }
    )


@pytest.fixture()
def falsy_names_df_dask(falsy_names_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(falsy_names_df_pandas, npartitions=2)


@pytest.fixture()
def falsy_names_df_spark(falsy_names_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(
        falsy_names_df_pandas.applymap(
            lambda tup: list(tup) if isinstance(tup, tuple) else tup
        )
    )


@pytest.fixture(
    params=["falsy_names_df_pandas", "falsy_names_df_dask", "falsy_names_df_spark"]
)
def falsy_names_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def sample_column_names(sample_df_pandas):
    return sample_df_pandas.columns.to_list()


@pytest.fixture()
def sample_inferred_logical_types():
    return {
        "id": Integer,
        "full_name": Unknown,
        "email": EmailAddress,
        "phone_number": PhoneNumber,
        "age": IntegerNullable,
        "signup_date": Datetime,
        "is_registered": BooleanNullable,
        "double": Double,
        "double_with_nan": Double,
        "integer": Integer,
        "nullable_integer": IntegerNullable,
        "boolean": Boolean,
        "categorical": Categorical,
        "datetime_with_NaT": Datetime,
        "url": URL,
        "ip_address": IPAddress,
    }


@pytest.fixture()
def sample_correct_logical_types():
    return {
        "id": Integer,
        "full_name": PersonFullName,
        "email": EmailAddress,
        "phone_number": PhoneNumber,
        "age": IntegerNullable,
        "signup_date": Datetime,
        "is_registered": BooleanNullable,
        "double": Double,
        "double_with_nan": Double,
        "integer": Integer,
        "nullable_integer": IntegerNullable,
        "boolean": Boolean,
        "categorical": Categorical,
        "datetime_with_NaT": Datetime,
        "url": URL,
        "ip_address": IPAddress,
    }


@pytest.fixture()
def serialize_df_pandas():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "cat_int": [1, 2, 1],
            "ord_int": [1, 2, 1],
            "cat_float": [1.0, 2.0, 1.0],
            "ord_float": [1.0, 2.0, 1.0],
            "cat_bool": [True, False, True],
            "ord_bool": [True, False, True],
        }
    )
    return df


@pytest.fixture()
def serialize_df_dask(serialize_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(serialize_df_pandas, npartitions=2)


@pytest.fixture()
def serialize_df_spark(serialize_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(serialize_df_pandas)


@pytest.fixture(
    params=["serialize_df_pandas", "serialize_df_dask", "serialize_df_spark"]
)
def serialize_df(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def pandas_datetimes():
    return [
        pd.Series(["3/11/2000", "3/12/2000", "3/13/2000", "3/14/2000"]),
        pd.Series(["3/11/2000", np.nan, "3/13/2000", "3/14/2000"]),
    ]


@pytest.fixture()
def dask_datetimes(pandas_datetimes):
    return [pd_to_dask(series) for series in pandas_datetimes]


@pytest.fixture()
def spark_datetimes(pandas_datetimes):
    return [pd_to_spark(series) for series in pandas_datetimes]


@pytest.fixture(params=["pandas_datetimes", "dask_datetimes", "spark_datetimes"])
def datetimes(request):
    return request.getfixturevalue(request.param)


@pytest.fixture()
def outliers_df_pandas():
    return pd.DataFrame(
        {
            "has_outliers": [93, 42, 37, -16, 49, 42, 36, 57, 60, 23],
            "no_outliers": [60, 42, 37, 23, 49, 42, 36, 57, 60, 23.0],
            "non_numeric": ["a"] * 10,
            "has_outliers_with_nans": [None, 42, 37, -16, 49, 93, 36, 57, 60, 23],
            "nans": pd.Series([None] * 10, dtype="float64"),
        }
    )


@pytest.fixture()
def outliers_df_dask(outliers_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(outliers_df_pandas, npartitions=2)


@pytest.fixture()
def outliers_df_spark(outliers_df_pandas):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(outliers_df_pandas)


@pytest.fixture(params=["outliers_df_pandas", "outliers_df_dask", "outliers_df_spark"])
def outliers_df(request):
    return request.getfixturevalue(request.param)


class MockCallback:
    def __init__(self):
        self.progress_history = []
        self.total_update = 0
        self.total_elapsed_time = 0

    def __call__(self, update, progress, total, unit, time_elapsed):
        self.total_update += update
        self.total = total
        self.progress_history.append(progress)
        self.unit = unit
        self.total_elapsed_time = time_elapsed


@pytest.fixture()
def mock_callback():
    return MockCallback()


class MockResultCallback:
    def __init__(self):
        self.results_so_far = []
        self.most_recent_calculation = []

    def __call__(self, results_so_far, most_recent_calculations):
        self.results_so_far.append(results_so_far)
        self.most_recent_calculation.append(most_recent_calculations)


@pytest.fixture()
def mock_results_callback():
    return MockResultCallback()


@pytest.fixture()
def timezones_df_pandas():
    dtypes = {
        "default_1": "datetime64[ns]",
        "utc_1": "datetime64[ns, UTC]",
        "eastern_1": "datetime64[ns, US/Eastern]",
        "default_2": "object",
        "utc_2": "object",
        "eastern_2": "object",
    }
    return pd.DataFrame(
        data=[
            [
                "2022-01-01",
                "2022-01-01 00:00:00+00:00",
                "2022-01-01 00:00:00-05:00",
                "2022-01-01",
                "2022-01-01T00:00:00+00:00",
                "2022-01-01 00:00:00-05:00",
            ],
            [
                "2022-01-02",
                "2022-01-02 00:00:00+00:00",
                "2022-01-02 00:00:00-05:00",
                "2022-01-02",
                "2022-01-02T00:00:00+00:00",
                "2022-01-02 00:00:00-05:00",
            ],
            ["NaT", "NaT", "NaT", "NaT", "NaT", "NaT"],
        ],
        columns=["default_1", "utc_1", "eastern_1", "default_2", "utc_2", "eastern_2"],
    ).astype(dtypes)


@pytest.fixture()
def timezones_df_dask(timezones_df_pandas):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(timezones_df_pandas, npartitions=2)


@pytest.fixture(params=["timezones_df_pandas", "timezones_df_dask"])
def timezones_df(request):
    return request.getfixturevalue(request.param)



from functools import wraps

import numpy as np
import pandas as pd

from woodwork.exceptions import ColumnNotPresentInSchemaError, WoodworkNotInitError
from woodwork.utils import _get_column_logical_type, import_or_none

dd = import_or_none("dask.dataframe")
ps = import_or_none("pyspark.pandas")


def init_series(
    series,
    logical_type=None,
    semantic_tags=None,
    use_standard_tags=True,
    description=None,
    origin=None,
    metadata=None,
):
    """Initializes Woodwork typing information for a series, numpy.ndarray or pd.api.extensions.
    ExtensionArray, returning a new Series. The dtype of the returned series will be converted
    to match the dtype associated with the LogicalType.

    Args:
        series (pd.Series, dd.Series, ps.Series, numpy.ndarray or pd.api.extensions.ExtensionArray):
            The original series from which to create the Woodwork initialized series.
        logical_type (LogicalType or str, optional): The logical type that should be assigned
            to the series. If no value is provided, the LogicalType for the series will
            be inferred.
        semantic_tags (str or list or set, optional): Semantic tags to assign to the series.
            Defaults to an empty set if not specified. There are two options for
            specifying the semantic tags:
            (str) If only one semantic tag is being set, a single string can be passed.
            (list or set) If multiple tags are being set, a list or set of strings can be passed.
        use_standard_tags (bool, optional): If True, will add standard semantic tags to the series
            based on the inferred or specified logical type of the series. Defaults to True.
        description (str, optional): Optional text describing the contents of the series.
        origin (str, optional): Optional text specifying origin of the column (i.e. "base" or "engineered").
        metadata (dict[str -> json serializable], optional): Metadata associated with the series.

    Returns:
        Series: A series with Woodwork typing information initialized
    """
    if not _is_series(series):
        if (
            isinstance(series, (np.ndarray, pd.api.extensions.ExtensionArray))
            and series.ndim == 1
        ):
            series = pd.Series(series)
        elif isinstance(series, np.ndarray) and series.ndim != 1:
            raise ValueError(
                f"np.ndarray input must be 1 dimensional. Current np.ndarray is {series.ndim} dimensional"
            )
        else:
            raise TypeError(
                f"Input must be of series type. The current input is of type {type(series)}"
            )
    logical_type = _get_column_logical_type(series, logical_type, series.name)
    new_series = logical_type.transform(series)
    new_series.ww.init(
        logical_type=logical_type,
        semantic_tags=semantic_tags,
        use_standard_tags=use_standard_tags,
        description=description,
        origin=origin,
        metadata=metadata,
    )
    return new_series


def _is_series(data):
    if isinstance(data, pd.Series):
        return True
    elif _is_dask_series(data):
        return True
    elif _is_spark_series(data):
        return True
    return False


def _is_dataframe(data):
    if isinstance(data, pd.DataFrame):
        return True
    elif _is_dask_dataframe(data):
        return True
    elif _is_spark_dataframe(data):
        return True
    return False


def get_invalid_schema_message(dataframe, schema):
    """Return a message indicating the reason that the provided schema cannot be used to
    initialize Woodwork on the dataframe. If the schema is valid for the dataframe,
    None will be returned.

    Args:
        dataframe (DataFrame): The dataframe against which to check the schema.
        schema (ww.TableSchema): The schema to use in the validity check.

    Returns:
        str or None: The reason that the schema is invalid for the dataframe
    """
    dataframe_cols = set(dataframe.columns)
    schema_cols = set(schema.columns.keys())

    df_cols_not_in_schema = dataframe_cols - schema_cols
    if df_cols_not_in_schema:
        return (
            f"The following columns in the DataFrame were missing from the typing information: "
            f"{df_cols_not_in_schema}"
        )
    schema_cols_not_in_df = schema_cols - dataframe_cols
    if schema_cols_not_in_df:
        return (
            f"The following columns in the typing information were missing from the DataFrame: "
            f"{schema_cols_not_in_df}"
        )
    logical_types = schema.logical_types
    for name in dataframe.columns:
        df_dtype = dataframe[name].dtype
        valid_dtype = logical_types[name]._get_valid_dtype(type(dataframe[name]))
        if str(df_dtype) != valid_dtype:
            return (
                f"dtype mismatch for column {name} between DataFrame dtype, "
                f"{df_dtype}, and {logical_types[name]} dtype, {valid_dtype}"
            )
    if schema.index is not None and isinstance(dataframe, pd.DataFrame):
        # Index validation not performed for Dask/Spark
        if not pd.Series(dataframe.index, dtype=dataframe[schema.index].dtype).equals(
            pd.Series(dataframe[schema.index].values)
        ):
            return "Index mismatch between DataFrame and typing information"
        elif not dataframe[schema.index].is_unique:
            return "Index column is not unique"
        elif dataframe[schema.index].isnull().any():
            return "Index contains null values"


def is_schema_valid(dataframe, schema):
    """Check if a schema is valid for initializing Woodwork on a dataframe

    Args:
        dataframe (DataFrame): The dataframe against which to check the schema.
        schema (ww.TableSchema): The schema to use in the validity check.

    Returns:
        boolean: Boolean indicating whether the schema is valid for the dataframe
    """

    invalid_schema_message = get_invalid_schema_message(dataframe, schema)
    if invalid_schema_message:
        return False
    return True


def _is_dask_series(data):
    if dd and isinstance(data, dd.Series):
        return True
    return False


def _is_dask_dataframe(data):
    if dd and isinstance(data, dd.DataFrame):
        return True
    return False


def _is_spark_dataframe(data):
    if ps and isinstance(data, ps.DataFrame):
        return True
    return False


def _is_spark_series(data):
    if ps and isinstance(data, ps.Series):
        return True
    return False


def _check_column_schema(method):
    """Decorator for WoodworkColumnAccessor that checks schema initialization"""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self._schema is None:
            msg = (
                "Woodwork not initialized for this Series. Initialize by "
                "calling Series.ww.init"
            )
            raise WoodworkNotInitError(msg)
        return method(self, *args, **kwargs)

    return wrapper


def _check_table_schema(method):
    """Decorator for WoodworkTableAccessor that checks schema initialization"""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if self._schema is None:
            msg = (
                "Woodwork not initialized for this DataFrame. Initialize by "
                "calling DataFrame.ww.init"
            )
            raise WoodworkNotInitError(msg)
        diff_cols = set(self._dataframe.columns).difference(
            set(self._schema.columns.keys())
        )
        if diff_cols:
            raise ColumnNotPresentInSchemaError(sorted(list(diff_cols)))
        return method(self, *args, **kwargs)

    return wrapper



"""
All code in this module was copied from pandas and is subject to the following
license:

BSD 3-Clause License

Copyright (c) 2008-2011, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team
All rights reserved.

Copyright (c) 2011-2021, Open source contributors.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import pytest

from woodwork.pandas_backport import guess_datetime_format


# TODO: Remove this test and also the
# `woodwork.pandas_backport.guess_datetime_format` backport when this fix is
# released in pandas 1.4:
# https://github.com/pandas-dev/pandas/pull/43900
@pytest.mark.parametrize(
    "string,fmt",
    [
        ("20111230", "%Y%m%d"),
        ("2011-12-30", "%Y-%m-%d"),
        ("30-12-2011", "%d-%m-%Y"),
        ("2011-12-30 00:00:00", "%Y-%m-%d %H:%M:%S"),
        ("2011-12-30T00:00:00", "%Y-%m-%dT%H:%M:%S"),
        ("2011-12-30T00:00:00UTC", "%Y-%m-%dT%H:%M:%S%Z"),
        ("2011-12-30T00:00:00Z", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00+9", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00+09", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00+090", None),
        ("2011-12-30T00:00:00+0900", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00-0900", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00+09:00", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00+09:000", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00+9:0", "%Y-%m-%dT%H:%M:%S%z"),
        ("2011-12-30T00:00:00+09:", None),
        ("2011-12-30T00:00:00.000000UTC", "%Y-%m-%dT%H:%M:%S.%f%Z"),
        ("2011-12-30T00:00:00.000000Z", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000+9", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000+09", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000+090", None),
        ("2011-12-30T00:00:00.000000+0900", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000-0900", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000+09:00", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000+09:000", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000+9:0", "%Y-%m-%dT%H:%M:%S.%f%z"),
        ("2011-12-30T00:00:00.000000+09:", None),
        ("2011-12-30 00:00:00.000000", "%Y-%m-%d %H:%M:%S.%f"),
        ("Tue 24 Aug 2021 01:30:48 AM", "%a %d %b %Y %H:%M:%S %p"),
        ("Tuesday 24 Aug 2021 01:30:48 AM", "%A %d %b %Y %H:%M:%S %p"),
    ],
)
def test_guess_datetime_format_with_parseable_formats(string, fmt):
    result = guess_datetime_format(string)
    assert result == fmt



import ast
import importlib
import re
from inspect import isclass
from mimetypes import add_type, guess_type
from timeit import default_timer as timer

import numpy as np
import pandas as pd

import woodwork as ww
from woodwork.exceptions import TypeValidationError
from woodwork.pandas_backport import guess_datetime_format

# Dictionary mapping formats/content types to the appropriate pandas read function
type_to_read_func_map = {
    "csv": pd.read_csv,
    "text/csv": pd.read_csv,
    "parquet": pd.read_parquet,
    "application/parquet": pd.read_parquet,
    "arrow": pd.read_feather,
    "application/arrow": pd.read_feather,
    "feather": pd.read_feather,
    "application/feather": pd.read_feather,
    "orc": pd.read_orc,
    "application/orc": pd.read_orc,
}

PYARROW_ERR_MSG = (
    "The pyarrow library is required to read from parquet/arrow/feather files.\n"
    "Install via pip:\n"
    "    pip install 'pyarrow>=3.0.0'\n"
    "Install via conda:\n"
    "    conda install 'pyarrow>=3.0.0'"
)

# Add new mimetypes
add_type("application/parquet", ".parquet")
add_type("application/arrow", ".arrow")
add_type("application/feather", ".feather")
add_type("application/orc", ".orc")


def import_or_none(library):
    """Attempts to import the requested library.

    Args:
        library (str): the name of the library
    Returns: the library if it is installed, else None
    """
    try:
        return importlib.import_module(library)
    except ImportError:
        return None


def camel_to_snake(s):
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def _convert_input_to_set(semantic_tags, error_language="semantic_tags", validate=True):
    """Takes input as a single string, a list of strings, or a set of strings
    and returns a set with the supplied values. If no values are supplied,
    an empty set will be returned."""
    if not semantic_tags:
        return set()

    if validate:
        _validate_tags_input_type(semantic_tags, error_language)

    if isinstance(semantic_tags, str):
        return {semantic_tags}

    if isinstance(semantic_tags, list):
        semantic_tags = set(semantic_tags)

    if validate:
        _validate_string_tags(semantic_tags, error_language)

    return semantic_tags


def _validate_tags_input_type(semantic_tags, error_language):
    if type(semantic_tags) not in [list, set, str]:
        raise TypeError(f"{error_language} must be a string, set or list")


def _validate_string_tags(semantic_tags, error_language):
    if not all([isinstance(tag, str) for tag in semantic_tags]):
        raise TypeError(f"{error_language} must contain only strings")


def read_file(
    filepath=None,
    content_type=None,
    name=None,
    index=None,
    time_index=None,
    semantic_tags=None,
    logical_types=None,
    use_standard_tags=True,
    column_origins=None,
    replace_nan=False,
    validate=True,
    **kwargs,
):
    """Read data from the specified file and return a DataFrame with initialized Woodwork typing information.

        Note:
            As the engine `fastparquet` cannot handle nullable pandas dtypes, `pyarrow` will be used
            for reading from parquet and arrow.

    Args:
        filepath (str): A valid string path to the file to read
        content_type (str): Content type of file to read
        name (str, optional): Name used to identify the DataFrame.
        index (str, optional): Name of the index column.
        time_index (str, optional): Name of the time index column.
        semantic_tags (dict, optional): Dictionary mapping column names in the dataframe to the
            semantic tags for the column. The keys in the dictionary should be strings
            that correspond to columns in the underlying dataframe. There are two options for
            specifying the dictionary values:
            (str): If only one semantic tag is being set, a single string can be used as a value.
            (list[str] or set[str]): If multiple tags are being set, a list or set of strings can be
            used as the value.
            Semantic tags will be set to an empty set for any column not included in the
            dictionary.
        logical_types (dict[str -> LogicalType], optional): Dictionary mapping column names in
            the dataframe to the LogicalType for the column. LogicalTypes will be inferred
            for any columns not present in the dictionary.
        use_standard_tags (bool, optional): If True, will add standard semantic tags to columns based
            on the inferred or specified logical type for the column. Defaults to True.
        column_origins (str or dict[str -> str], optional): Origin of each column. If a string is supplied, it is
                used as the origin for all columns. A dictionary can be used to set origins for individual columns.
        replace_nan (bool, optional): Whether to replace empty string values and string representations of
            NaN values ("nan", "<NA>") with np.nan or pd.NA values based on column dtype. Defaults to False.
        validate (bool, optional): Whether parameter and data validation should occur. Defaults to True. Warning:
                Should be set to False only when parameters and data are known to be valid.
                Any errors resulting from skipping validation with invalid inputs may not be easily understood.
        **kwargs: Additional keyword arguments to pass to the underlying pandas read file function. For more
            information on available keywords refer to the pandas documentation.

    Returns:
        pd.DataFrame: DataFrame created from the specified file with Woodwork typing information initialized.
    """
    if content_type is None:
        inferred_type, _ = guess_type(filepath)
        if inferred_type is None:
            raise RuntimeError(
                "Content type could not be inferred. Please specify content_type and try again."
            )
        content_type = inferred_type

    if content_type not in type_to_read_func_map:
        raise RuntimeError(
            "Reading from content type {} is not currently supported".format(
                content_type
            )
        )

    pyarrow_types = [
        "parquet",
        "application/parquet",
        "arrow",
        "application/arrow",
        "feather",
        "application/feather",
        "orc",
        "application/orc",
    ]
    if content_type in pyarrow_types:
        import_or_raise("pyarrow", PYARROW_ERR_MSG)
        if content_type in ["parquet", "application/parquet"]:
            kwargs["engine"] = "pyarrow"

    dataframe = type_to_read_func_map[content_type](filepath, **kwargs)

    if replace_nan:
        dataframe = _replace_nan_strings(dataframe)

    dataframe.ww.init(
        name=name,
        index=index,
        time_index=time_index,
        semantic_tags=semantic_tags,
        logical_types=logical_types,
        use_standard_tags=use_standard_tags,
        column_origins=column_origins,
        validate=validate,
    )
    return dataframe


def import_or_raise(library, error_msg):
    """Attempts to import the requested library.  If the import fails, raises an
    ImportError with the supplied error message.

    Args:
        library (str): the name of the library
        error_msg (str): error message to return if the import fails
    """
    try:
        return importlib.import_module(library)
    except ImportError:
        raise ImportError(error_msg)


def _is_s3(string):
    """Checks if the given string is a s3 path. Returns a boolean."""
    return "s3://" in string


def _is_url(string):
    """Checks if the given string is an url path. Returns a boolean."""
    return "http" in string


def _reformat_to_latlong(latlong, is_spark=False):
    """
    Accepts 2-tuple like values, or a single NaN like value.
    NaN like values are replaced with np.nan.
    """
    if isinstance(latlong, str):

        nan_values = [x for x in ww.config.get_option("nan_values") if len(x)]
        nan_values = "|".join(nan_values)

        latlong = re.sub(nan_values, "None", latlong)

        try:
            latlong = ast.literal_eval(latlong)
        except ValueError:
            pass

    if isinstance(latlong, (list, tuple)):
        if len(latlong) != 2:
            raise TypeValidationError(
                f"LatLong values must have exactly two values. {latlong} does not have two values."
            )

        latitude, longitude = latlong

        try:
            latitude = _coerce_to_float(latitude)
            longitude = _coerce_to_float(longitude)
        except ValueError:
            raise TypeValidationError(
                f"LatLong values must be in decimal degrees. {latlong} does not have latitude or longitude values that can be converted to a float."
            )

        if is_spark:
            return [latitude, longitude]
        return (latitude, longitude)

    if _is_nan(latlong):
        return np.nan

    raise TypeValidationError(
        f"""LatLong value is not properly formatted. Value must be one of the following:
- A 2-tuple or list of 2 values representing decimal latitude or longitude values (NaN values are allowed).
- A single NaN value.
- A string representation of the above.

{latlong} does not fit the criteria."""
    )


def _coerce_to_float(val):
    """Attempts to convert a value to a float, propagating null values."""
    if _is_nan(val):
        return np.nan

    try:
        return float(val)
    except (ValueError, TypeError):
        raise ValueError(
            f"The value represented by {val} cannot be converted to a float."
        )


def _is_valid_latlong_series(series):
    """Returns True if all elements in the series contain properly formatted LatLong values,
    otherwise returns False"""
    if ww.accessor_utils._is_dask_series(series):
        series = series.get_partition(0).compute()
    if ww.accessor_utils._is_spark_series(series):
        series = series.to_pandas()
        is_spark = True
    else:
        is_spark = False
    if series.apply(_is_valid_latlong_value, args=(is_spark,)).all():
        return True
    return False


def _is_valid_latlong_value(val, is_spark=False):
    """Returns True if the value provided is a properly formatted LatLong value for a
    pandas, Dask or Spark Series, otherwise returns False."""
    if isinstance(val, (list, tuple)):
        if len(val) != 2:
            return False

        if not isinstance(val, list if is_spark else tuple):
            return False

        latitude, longitude = val
        return isinstance(latitude, float) and isinstance(longitude, float)

    if isinstance(val, float):
        return np.isnan(val)

    if is_spark and val is None:
        return True

    return False


def _is_nan(value):
    """This function checks if string values are common NaN values.
    Lists are not counted as NaN, and all other values are passed to pd.isnull
    """
    if isinstance(value, str):
        return value in ww.config.get_option("nan_values")
    if isinstance(value, list):
        return False
    return pd.isnull(value)


def _is_latlong_nan(value):
    """Checks if a LatLong value is NaN"""
    if isinstance(value, (tuple, list)):
        return all([_is_nan(x) for x in value])

    return _is_nan(value)


def get_valid_mi_types():
    """
    Generate a list of LogicalTypes that are valid for calculating mutual information. Note that
    index columns are not valid for calculating mutual information, but their types may be
    returned by this function.

    Args:
        None

    Returns:
        list(LogicalType): A list of the LogicalTypes that can be use to calculate mutual information
    """
    valid_types = []
    for ltype in ww.type_system.registered_types:
        if "category" in ltype.standard_tags:
            valid_types.append(ltype)
        elif "numeric" in ltype.standard_tags:
            valid_types.append(ltype)
        elif (
            ltype == ww.logical_types.Datetime
            or ltype == ww.logical_types.Boolean
            or ltype == ww.logical_types.BooleanNullable
        ):
            valid_types.append(ltype)

    return valid_types


def get_valid_pearson_types():
    """
    Generate a list of LogicalTypes that are valid for calculating Pearson correlation. Note that
    index columns are not valid for calculating dependence, but their types may be
    returned by this function.

    Args:
        None

    Returns:
        list(LogicalType): A list of the LogicalTypes that can be use to calculate Pearson correlation
    """
    valid_types = []
    for ltype in ww.type_system.registered_types:
        if "numeric" in ltype.standard_tags:
            valid_types.append(ltype)
        elif ltype == ww.logical_types.Datetime:
            valid_types.append(ltype)

    return valid_types


def _get_column_logical_type(series, logical_type, name):
    if logical_type:
        return _parse_logical_type(logical_type, name)
    else:
        return ww.type_system.infer_logical_type(series)


def _parse_logical_type(logical_type, name):
    if isinstance(logical_type, str):
        logical_type = ww.type_system.str_to_logical_type(logical_type)

    if isclass(logical_type):
        logical_type = logical_type()

    if type(logical_type) not in ww.type_system.registered_types:
        raise TypeError(f"Invalid logical type specified for '{name}'")

    return logical_type


def concat_columns(objs, validate_schema=True):
    """
    Concatenate Woodwork objects along the columns axis. There can only be one index and time index
    set across the objects passed in. As Woodwork does not allow duplicate column names,
    will not allow duplicate columns at concatenation.

    Args:
        objs (list[Series, DataFrame]): The Woodwork objects to be concatenated. If Woodwork
            is not initialized on any of the objects, type inference will be performed.
        validate_schema (bool, optional): Whether validation should be performed on the typing information
            for the concatenated DataFrame. Defaults to True.

    Returns:
        DataFrame: A Woodwork dataframe whose typing information is also a concatenation of the input dataframes.
    """
    if not objs:
        raise ValueError("No objects to concatenate")

    table_name = ""

    logical_types = {}
    semantic_tags = {}
    col_descriptions = {}
    col_origins = {}
    col_metadata = {}
    table_metadata = {}
    use_standard_tags = {}

    index = None
    time_index = None

    # Record the typing information for all the columns that have Woodwork schemas
    col_names_seen = set()
    for obj in objs:
        ww_columns = {}
        if isinstance(obj.ww.schema, ww.table_schema.TableSchema):
            # Raise error if there's overlap between table metadata
            overlapping_keys = obj.ww.metadata.keys() & table_metadata.keys()
            if overlapping_keys:
                raise ValueError(
                    f"Cannot resolve overlapping keys in table metadata: {overlapping_keys}"
                )

            table_metadata = {**obj.ww.metadata, **table_metadata}

            # Combine table names
            if obj.ww.name is not None:
                if table_name:
                    table_name += "_"
                table_name += str(obj.ww.name)

            # Cannot have multiple tables with indexes or time indexes set
            if obj.ww.index is not None:
                if index is None:
                    index = obj.ww.index
                else:
                    raise IndexError(
                        "Cannot set the Woodwork index of multiple input objects. "
                        "Please remove the index columns from all but one table."
                    )
            if obj.ww.time_index is not None:
                if time_index is None:
                    time_index = obj.ww.time_index
                else:
                    raise IndexError(
                        "Cannot set the Woodwork time index of multiple input objects. "
                        "Please remove the time index columns from all but one table."
                    )

            ww_columns = obj.ww.schema.columns
        elif isinstance(obj.ww.schema, ww.column_schema.ColumnSchema):
            ww_columns = {obj.name: obj.ww.schema}

        # Compile the typing information per column
        for name, col_schema in ww_columns.items():
            if name in col_names_seen:
                raise ValueError(
                    f"Duplicate column '{name}' has been found in more than one input object. "
                    "Please remove duplicate columns from all but one table."
                )
            logical_types[name] = col_schema.logical_type
            semantic_tags[name] = col_schema.semantic_tags - {"time_index"} - {"index"}
            col_metadata[name] = col_schema.metadata
            col_descriptions[name] = col_schema.description
            col_origins[name] = col_schema.origin
            use_standard_tags[name] = col_schema.use_standard_tags

            col_names_seen.add(name)

    # Perform concatenation with the correct library
    obj = objs[0]
    dd = import_or_none("dask.dataframe")
    ps = import_or_none("pyspark.pandas")

    lib = pd
    if ww.accessor_utils._is_spark_dataframe(obj) or ww.accessor_utils._is_spark_series(
        obj
    ):
        lib = ps
    elif ww.accessor_utils._is_dask_dataframe(obj) or ww.accessor_utils._is_dask_series(
        obj
    ):
        lib = dd

    combined_df = lib.concat(objs, axis=1, join="outer")

    # Initialize Woodwork with all of the typing information from the input objs
    # performing type inference on any columns that did not already have Woodwork initialized
    combined_df.ww.init(
        name=table_name or None,
        index=index,
        time_index=time_index,
        logical_types=logical_types,
        semantic_tags=semantic_tags,
        table_metadata=table_metadata or None,
        column_metadata=col_metadata,
        column_descriptions=col_descriptions,
        column_origins=col_origins,
        use_standard_tags=use_standard_tags,
        validate=validate_schema,
    )
    return combined_df


class CallbackCaller:
    """
    Helper class for updating progress of a function and making a call to the progress callback
    function, if provided. Adds the progress increment to the current progress.

    If provided, the callback function should accept the following parameters:
        - update (int): change in progress since last call
        - progress (int): the progress so far in the calculations
        - total (int): the total number of calculations to do
        - unit (str): unit of measurement for progress/total
        - time_elapsed (float): total time in seconds elapsed since start of call

    """

    def __init__(self, callback, unit, total, start_time=None, start_progress=0):
        """
        Args:
            callback (func): callback method to call
            unit (str): unit of measurement for progress/total
            total (int): the total number of calculations to do
            start_time (datetime): when time started for the callback.  Defaults
                to when the class instance is created
            start_progress (int): starting progress for the callback.  Defaults to 0.
        """
        if start_time is not None:
            self.start_time = start_time
        else:
            self.start_time = timer()
        self.callback = callback
        self.unit = unit
        self.current_progress = start_progress
        self.total = total

    def update(self, progress_increment):
        """
        Args:
            progress_increment (int): change in progress since the last call
        """
        if self.callback is not None:
            elapsed_time = timer() - self.start_time
            new_progress = self.current_progress + progress_increment
            self.callback(
                progress_increment, new_progress, self.total, self.unit, elapsed_time
            )
            self.current_progress = new_progress


def _infer_datetime_format(dates, n=100):
    """Helper function to infer the datetime format of the first n non-null rows of a series
    Args:
        dates (Series): Series of string or datetime string to guess the format of
        n (int): the maximum number of nonnull rows to sample from the series
    """
    first_n = dates.dropna().head(n)

    ps = import_or_none("pyspark.pandas")
    if ps and isinstance(first_n, ps.series.Series):
        first_n = first_n.to_pandas()

    if len(first_n) == 0:
        return None
    try:
        fmts = first_n.map(guess_datetime_format)
        mode_fmt = fmts.mode().loc[0]  # select first most common format
    except (TypeError, ValueError, IndexError, KeyError, NotImplementedError):
        mode_fmt = None
    return mode_fmt


def _replace_nan_strings(df: pd.DataFrame) -> pd.DataFrame:
    """Replaces empty string values and string representations of
    NaN values ("nan", "<NA>") with np.nan or pd.NA depending on
    column dtype."""
    df = df.fillna(value=np.nan)

    for col, dtype in df.dtypes.items():
        replace_val = np.nan
        if str(dtype) == "boolean":
            # All replace calls below fail with boolean dtype
            # but boolean cols cannot contain strings to begin with.
            continue
        elif str(dtype) == "string":
            # Must use pd.NA as replacement value for string dtype
            replace_val = pd.NA

        replaced_series = df[col].replace(r"^\s*$", replace_val, regex=True)
        replaced_series = replaced_series.replace(
            ww.config.get_option("nan_values"), replace_val
        )
        df[col] = replaced_series

    return df
